from q3as.algo.vqe import VQE, HaltReason, VQEIteration, QAOA

import numpy as np
from qiskit_aer import AerSimulator
from qiskit_aer.primitives import EstimatorV2 as Estimator, SamplerV2 as Sampler
from qiskit import transpile
from qiskit.circuit.library import EfficientSU2
from qiskit.quantum_info import SparsePauliOp

simulator = AerSimulator()


def test_vqe_with_init_params():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    ansatz = EfficientSU2(hamiltonian.num_qubits)
    x0 = 2 * np.pi * np.random.random(ansatz.num_parameters)
    vqe = (
        VQE.builder()
        .ansatz(transpile(ansatz, simulator))
        .observables(hamiltonian)
        .initial_params(x0)
        .build()
    )
    res = vqe.run(Estimator())
    assert res.reason == HaltReason.TOLERANCE


def test_vqe_no_init_params():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    ansatz = EfficientSU2(hamiltonian.num_qubits)
    vqe = (
        VQE.builder()
        .ansatz(transpile(ansatz, simulator))
        .observables(hamiltonian)
        .build()
    )
    res = vqe.run(Estimator())
    assert res.reason == HaltReason.TOLERANCE


def test_vqe_with_maxiter():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    ansatz = EfficientSU2(hamiltonian.num_qubits)
    vqe = (
        VQE.builder()
        .ansatz(transpile(ansatz, simulator))
        .observables(hamiltonian)
        .maxiter(100)
        .build()
    )
    res = vqe.run(Estimator())
    assert res.reason == HaltReason.MAXITER


def test_vqe_with_interrupt():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    ansatz = EfficientSU2(hamiltonian.num_qubits)
    vqe = (
        VQE.builder()
        .ansatz(transpile(ansatz, simulator))
        .observables(hamiltonian)
        .maxiter(100)
        .build()
    )

    def callback(iter: VQEIteration):
        if iter.iter >= 10:
            raise StopIteration

    res = vqe.run(Estimator(), callback=callback)
    assert res.reason == HaltReason.INTERRUPT


def test_vqe_with_callable_ansatz():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    res = (
        VQE.builder()
        .ansatz(QAOA(lambda qc: transpile(qc, simulator)))
        .observables(hamiltonian)
        .run(Estimator())
    )
    assert res.reason == HaltReason.TOLERANCE


def test_vqe_with_sampler():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    res = (
        VQE.builder()
        .ansatz(QAOA(lambda qc: transpile(qc, simulator)))
        .observables(hamiltonian)
        .run(Estimator(), Sampler())
    )
    assert res.reason == HaltReason.TOLERANCE



from .problems import (
    QuadraticProgram,
    LinearConstraint,
    QuadraticConstraint,
    Variable,
    LinearExpression,
    QuadraticExpression,
    QuadraticObjective,
    VarType,
    QuadraticProgramStatus,
    QuadraticProgramElement,
    Constraint,
)
from .converters import QuadraticProgramToQubo
from .translators import from_ising, to_ising

__all__ = [
    "QuadraticProgram",
    "QuadraticProgramToQubo",
    "QuadraticProgramElement",
    "VarType",
    "Variable",
    "LinearExpression",
    "QuadraticExpression",
    "Constraint",
    "LinearConstraint",
    "QuadraticConstraint",
    "QuadraticObjective",
    "QuadraticProgramStatus",
    "from_ising",
    "to_ising",
]



from .vqe import VQE, VQEBuilder, VQEIteration, VQEResult, QAOA, TwoLocal, EfficientSU2
from .optimizer import Optimizer, COBYLA, SLSQP
from .types import HaltReason

__all__ = [
    "VQE",
    "VQEBuilder",
    "VQEIteration",
    "VQEResult",
    "QAOA",
    "TwoLocal",
    "EfficientSU2",
    "Optimizer",
    "COBYLA",
    "SLSQP",
    "HaltReason",
]



import os

from q3as.algo.vqe import VQE, EfficientSU2
from q3as.app import Maxcut
from q3as import Client, Credentials

from qiskit_aer import AerSimulator
from qiskit_aer.primitives import EstimatorV2 as Estimator, SamplerV2 as Sampler

from qiskit import transpile

url = os.getenv("Q3AS_URL", "http://localhost:8080")

simulator = AerSimulator()


def simple_maxcut():
    return Maxcut(
        [
            (0, 1, 1.0),
            (0, 2, 1.0),
            (0, 4, 1.0),
            (1, 2, 1.0),
            (2, 3, 1.0),
            (3, 4, 1.0),
        ]
    )


def test_maxcut():
    _ = (
        VQE.builder()
        .app(simple_maxcut())
        .ansatz(EfficientSU2(lambda c: transpile(c, simulator)))
        .run(Estimator(), Sampler())
    )


def test_client_quadratic():
    with Client(Credentials.load(".credentials.json"), url=url) as client:
        _ = VQE.builder().app(simple_maxcut()).send(client)



import pytest
import os
import random

from q3as.algo.vqe import VQE
from q3as.app import Maxcut
from q3as import Client, Credentials

from qiskit_aer import AerSimulator

simulator = AerSimulator()

benchmark = pytest.mark.skipif("not config.getoption('benchmark')")

url = os.getenv("Q3AS_URL", "http://localhost:8080")


def gen_graph(n, p=0.5, weight_range=(1, 1)):
    """
    Generates a random graph using the Erdős–Rényi model (G(n, p)).

    Parameters:
        n (int): Number of nodes.
        p (float): Probability of adding an edge between any two nodes.
        weight_range (tuple): Range (min, max) for the random weights of edges.

    Returns:
        edges (list of tuples): A list of edges represented as (node1, node2, weight).
    """
    edges = []

    # Iterate through all pairs of nodes (i, j) where i < j
    for i in range(n):
        for j in range(i + 1, n):
            # Include an edge with probability `p`
            if random.random() < p:
                # Generate a random weight within the specified range
                weight = random.randint(weight_range[0], weight_range[1])
                edges.append((i, j, weight))

    return edges


@benchmark
def test_benchmark():
    with Client(Credentials.load(".credentials.json"), url=url) as client:
        for size in [15, 20, 25, 28, 30, 32]:
            print("Running Maxcut with size", size)
            job = VQE.builder().app(Maxcut(gen_graph(size))).send(client)
            print("Started job with name:", job.name)
            print("Result", job.result())



from __future__ import annotations
from typing import Tuple, Any

import numpy as np

from q3as.quadratic import QuadraticProgram
from q3as.quadratic.converters import QuadraticProgramToQubo
from q3as.encoding.quadratic import EncodedQuadraticProgram

from .application import Application, ApplicationName, BitString


class Qubo(Application[EncodedQuadraticProgram, frozenset[Tuple[str, float]]]):
    def __init__(self, qp: QuadraticProgram):
        """
        Create a new QUBO application.
        """
        self.program = qp
        self.converter = QuadraticProgramToQubo()
        self.converted = self.converter.convert(qp)

    def name(self) -> ApplicationName:
        return "qubo"

    def encode(self) -> EncodedQuadraticProgram:
        return EncodedQuadraticProgram.encode(self.program)

    @classmethod
    def validate_encoded(cls, encoded: Any) -> EncodedQuadraticProgram:
        return EncodedQuadraticProgram.model_validate(encoded)

    @classmethod
    def decode(cls, encoded: EncodedQuadraticProgram) -> Qubo:
        return Qubo(EncodedQuadraticProgram.decode(encoded))

    def hamiltonian(self):
        """
        Return the Hamiltonian of the QUBO problem.
        """
        return self.converted.to_ising()[0]

    def interpret(self, bit_string: BitString) -> frozenset[Tuple[str, float]]:
        """
        Interpret the bit string as a solution to the QUBO problem. This will return a list of pairs of variable names and their assigned values
        """
        values = self.converter.interpret(np.flip(bit_string.to_list()))
        return frozenset(
            [
                (name, float(values[index]))
                for name, index in self.program.variables_index.items()
            ]
        )



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Quadratic Program."""

import logging
from collections.abc import Sequence
from math import isclose
from typing import Dict, List, Optional, Tuple, Union, cast
from warnings import warn

import numpy as np
from numpy import ndarray
from qiskit.quantum_info import SparsePauliOp
from qiskit.quantum_info.operators.base_operator import BaseOperator
from scipy.sparse import spmatrix

from ..exceptions import QiskitOptimizationError
from ..infinity import INFINITY
from .constraint import Constraint, ConstraintSense
from .linear_constraint import LinearConstraint
from .quadratic_constraint import QuadraticConstraint
from .quadratic_objective import QuadraticObjective
from .quadratic_program_element import QuadraticProgramElement
from .variable import Variable
from .types import QuadraticProgramStatus, VarType

logger = logging.getLogger(__name__)


class QuadraticProgram:
    """Quadratically Constrained Quadratic Program representation.

    This representation supports inequality and equality constraints,
    as well as continuous, binary, and integer variables.
    """

    Status = QuadraticProgramStatus

    def __init__(self, name: str = "") -> None:
        """
        Args:
            name: The name of the quadratic program.
        """
        if not name.isprintable():
            warn("Problem name is not printable")
        self._name = ""
        self.name = name
        self._status = QuadraticProgram.Status.VALID

        self._variables: List[Variable] = []
        self._variables_index: Dict[str, int] = {}

        self._linear_constraints: List[LinearConstraint] = []
        self._linear_constraints_index: Dict[str, int] = {}

        self._quadratic_constraints: List[QuadraticConstraint] = []
        self._quadratic_constraints_index: Dict[str, int] = {}

        self._objective = QuadraticObjective(self)

    def __repr__(self) -> str:
        from ..translators.prettyprint import DEFAULT_TRUNCATE, expr2str

        objective = expr2str(
            constant=self._objective.constant,
            linear=self.objective.linear,
            quadratic=self._objective.quadratic,
            truncate=DEFAULT_TRUNCATE,
        )
        num_constraints = (
            self.get_num_linear_constraints() + self.get_num_quadratic_constraints()
        )
        return (
            f"<{self.__class__.__name__}: "
            f"{self.objective.sense.name.lower()} "
            f"{objective}, "
            f"{self.get_num_vars()} variables, "
            f"{num_constraints} constraints, "
            f"'{self._name}'>"
        )

    def __str__(self) -> str:
        num_constraints = (
            self.get_num_linear_constraints() + self.get_num_quadratic_constraints()
        )
        return (
            f"{str(self.objective)} "
            f"({self.get_num_vars()} variables, "
            f"{num_constraints} constraints, "
            f"'{self._name}')"
        )

    def clear(self) -> None:
        """Clears the quadratic program, i.e., deletes all variables, constraints, the
        objective function as well as the name.
        """
        self._name = ""
        self._status = QuadraticProgram.Status.VALID

        self._variables.clear()
        self._variables_index.clear()

        self._linear_constraints.clear()
        self._linear_constraints_index.clear()

        self._quadratic_constraints.clear()
        self._quadratic_constraints_index.clear()

        self._objective = QuadraticObjective(self)

    @property
    def name(self) -> str:
        """Returns the name of the quadratic program.

        Returns:
            The name of the quadratic program.
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """Sets the name of the quadratic program.

        Args:
            name: The name of the quadratic program.
        """
        self._check_name(name, "Problem")
        self._name = name

    @property
    def status(self) -> QuadraticProgramStatus:
        """Status of the quadratic program.
        It can be infeasible due to variable substitution.

        Returns:
            The status of the quadratic program
        """
        return self._status

    @property
    def variables(self) -> List[Variable]:
        """Returns the list of variables of the quadratic program.

        Returns:
            List of variables.
        """
        return self._variables

    @property
    def variables_index(self) -> Dict[str, int]:
        """Returns the dictionary that maps the name of a variable to its index.

        Returns:
            The variable index dictionary.
        """
        return self._variables_index

    def _add_variable(
        self,
        lowerbound: Union[float, int],
        upperbound: Union[float, int],
        vartype: VarType,
        name: Optional[str],
    ) -> Variable:
        if not name:
            name = "x"
            key_format = "{}"
        else:
            key_format = ""
        return self._add_variables(
            1, lowerbound, upperbound, vartype, name, key_format
        )[1][0]

    def _add_variables(
        self,
        keys: Union[int, Sequence],
        lowerbound: Union[float, int],
        upperbound: Union[float, int],
        vartype: VarType,
        name: Optional[str],
        key_format: str,
    ) -> Tuple[List[str], List[Variable]]:
        if isinstance(keys, int) and keys < 1:
            raise QiskitOptimizationError(
                f"Cannot create non-positive number of variables: {keys}"
            )
        if not name:
            name = "x"
        if "{{}}" in key_format:
            raise QiskitOptimizationError(
                f"Formatter cannot contain nested substitutions: {key_format}"
            )
        if key_format.count("{}") > 1:
            raise QiskitOptimizationError(
                f"Formatter cannot contain more than one substitution: {key_format}"
            )

        def _find_name(name, key_format, k):
            prev = None
            while True:
                new_name = name + key_format.format(k)
                if new_name == prev:
                    raise QiskitOptimizationError(
                        f"Variable name already exists: {new_name}"
                    )
                if new_name in self._variables_index:
                    k += 1
                    prev = new_name
                else:
                    break
            return new_name, k + 1

        names = []
        variables = []
        k = self.get_num_vars()
        lst = keys if isinstance(keys, Sequence) else range(keys)
        for key in lst:
            if isinstance(keys, Sequence):
                indexed_name = name + key_format.format(key)
            else:
                indexed_name, k = _find_name(name, key_format, k)
            if indexed_name in self._variables_index:
                raise QiskitOptimizationError(
                    f"Variable name already exists: {indexed_name}"
                )
            self._check_name(indexed_name, "Variable")
            names.append(indexed_name)
            self._variables_index[indexed_name] = self.get_num_vars()
            variable = Variable(self, indexed_name, lowerbound, upperbound, vartype)
            self._variables.append(variable)
            variables.append(variable)
        return names, variables

    def _var_dict(
        self,
        keys: Union[int, Sequence],
        lowerbound: Union[float, int],
        upperbound: Union[float, int],
        vartype: VarType,
        name: Optional[str],
        key_format: str,
    ) -> Dict[str, Variable]:
        """
        Adds a positive number of variables to the variable list and index and returns a
        dictionary mapping the variable names to their instances. If 'key_format' is present,
        the next 'var_count' available indices are substituted into 'key_format' and appended
        to 'name'.

        Args:
            lowerbound: The lower bound of the variable(s).
            upperbound: The upper bound of the variable(s).
            vartype: The type of the variable(s).
            name: The name(s) of the variable(s).
            key_format: The format used to name/index the variable(s).
            keys: If keys: int, it is interpreted as the number of variables to construct.
                  Otherwise, the elements of the sequence are converted to strings via 'str' and
                  substituted into `key_format`.

        Returns:
            A dictionary mapping the variable names to variable instances.

        Raises:
            QiskitOptimizationError: if the variable name is already taken.
            QiskitOptimizationError: if less than one variable instantiation is attempted.
            QiskitOptimizationError: if `key_format` has more than one substitution or a
                                     nested substitution.
        """
        return dict(
            zip(
                *self._add_variables(
                    keys, lowerbound, upperbound, vartype, name, key_format
                )
            )
        )

    def _var_list(
        self,
        keys: Union[int, Sequence],
        lowerbound: Union[float, int],
        upperbound: Union[float, int],
        vartype: VarType,
        name: Optional[str],
        key_format: str,
    ) -> List[Variable]:
        """
        Adds a positive number of variables to the variable list and index and returns a
        list of variable instances.

        Args:
            lowerbound: The lower bound of the variable(s).
            upperbound: The upper bound of the variable(s).
            vartype: The type of the variable(s).
            name: The name(s) of the variable(s).
            key_format: The format used to name/index the variable(s).
            keys: If keys: int, it is interpreted as the number of variables to construct.
                  Otherwise, the elements of the sequence are converted to strings via 'str' and
                  substituted into `key_format`.

        Returns:
            A dictionary mapping the variable names to variable instances.

        Raises:
            QiskitOptimizationError: if the variable name is already taken.
            QiskitOptimizationError: if less than one variable instantiation is attempted.
            QiskitOptimizationError: if `key_format` has more than one substitution or a
                                     nested substitution.
        """
        return self._add_variables(
            keys, lowerbound, upperbound, vartype, name, key_format
        )[1]

    def continuous_var(
        self,
        lowerbound: Union[float, int] = 0,
        upperbound: Union[float, int] = INFINITY,
        name: Optional[str] = None,
    ) -> Variable:
        """Adds a continuous variable to the quadratic program.

        Args:
            lowerbound: The lowerbound of the variable.
            upperbound: The upperbound of the variable.
            name: The name of the variable.
                If it's ``None`` or empty ``""``, the default name, e.g., ``x0``, is used.

        Returns:
            The added variable.

        Raises:
            QiskitOptimizationError: if the variable name is already occupied.
        """
        return self._add_variable(
            lowerbound, upperbound, Variable.Type.CONTINUOUS, name
        )

    def continuous_var_dict(
        self,
        keys: Union[int, Sequence],
        lowerbound: Union[float, int] = 0,
        upperbound: Union[float, int] = INFINITY,
        name: Optional[str] = None,
        key_format: str = "{}",
    ) -> Dict[str, Variable]:
        """
        Uses 'var_dict' to construct a dictionary of continuous variables

        Args:
            lowerbound: The lower bound of the variable(s).
            upperbound: The upper bound of the variable(s).
            name: The name(s) of the variable(s).
                If it's ``None`` or empty ``""``, the default name, e.g., ``x0``, is used.
            key_format: The format used to name/index the variable(s).
            keys: If keys: int, it is interpreted as the number of variables to construct.
                  Otherwise, the elements of the sequence are converted to strings via 'str' and
                  substituted into `key_format`.

        Returns:
            A dictionary mapping the variable names to variable instances.

        Raises:
            QiskitOptimizationError: if the variable name is already taken.
            QiskitOptimizationError: if less than one variable instantiation is attempted.
            QiskitOptimizationError: if `key_format` has more than one substitution or a
                                     nested substitution.
        """
        return self._var_dict(
            keys, lowerbound, upperbound, Variable.Type.CONTINUOUS, name, key_format
        )

    def continuous_var_list(
        self,
        keys: Union[int, Sequence],
        lowerbound: Union[float, int] = 0,
        upperbound: Union[float, int] = INFINITY,
        name: Optional[str] = None,
        key_format: str = "{}",
    ) -> List[Variable]:
        """
        Uses 'var_list' to construct a list of continuous variables

        Args:
            lowerbound: The lower bound of the variable(s).
            upperbound: The upper bound of the variable(s).
            name: The name(s) of the variable(s).
                If it's ``None`` or empty ``""``, the default name, e.g., ``x0``, is used.
            key_format: The format used to name/index the variable(s).
            keys: If keys: int, it is interpreted as the number of variables to construct.
                  Otherwise, the elements of the sequence are converted to strings via 'str' and
                  substituted into `key_format`.

        Returns:
            A list of variable instances.

        Raises:
            QiskitOptimizationError: if the variable name is already taken.
            QiskitOptimizationError: if less than one variable instantiation is attempted.
            QiskitOptimizationError: if `key_format` has more than one substitution or a
                                     nested substitution.
        """
        return self._var_list(
            keys, lowerbound, upperbound, Variable.Type.CONTINUOUS, name, key_format
        )

    def binary_var(self, name: Optional[str] = None) -> Variable:
        """Adds a binary variable to the quadratic program.

        Args:
            name: The name of the variable.
                If it's ``None`` or empty ``""``, the default name, e.g., ``x0``, is used.

        Returns:
            The added variable.

        Raises:
            QiskitOptimizationError: if the variable name is already occupied.
        """
        return self._add_variable(0, 1, Variable.Type.BINARY, name)

    def binary_var_dict(
        self,
        keys: Union[int, Sequence],
        name: Optional[str] = None,
        key_format: str = "{}",
    ) -> Dict[str, Variable]:
        """
        Uses 'var_dict' to construct a dictionary of binary variables

        Args:
            name: The name(s) of the variable(s).
                If it's ``None`` or empty ``""``, the default name, e.g., ``x0``, is used.
            key_format: The format used to name/index the variable(s).
            keys: If keys: int, it is interpreted as the number of variables to construct.
                  Otherwise, the elements of the sequence are converted to strings via 'str' and
                  substituted into `key_format`.

        Returns:
            A dictionary mapping the variable names to variable instances.

        Raises:
            QiskitOptimizationError: if the variable name is already taken.
            QiskitOptimizationError: if less than one variable instantiation is attempted.
            QiskitOptimizationError: if `key_format` has more than one substitution or a
                                     nested substitution.
        """
        return self._var_dict(keys, 0, 1, Variable.Type.BINARY, name, key_format)

    def binary_var_list(
        self,
        keys: Union[int, Sequence],
        name: Optional[str] = None,
        key_format: str = "{}",
    ) -> List[Variable]:
        """
        Uses 'var_list' to construct a list of binary variables

        Args:
            name: The name(s) of the variable(s).
                If it's ``None`` or empty ``""``, the default name, e.g., ``x0``, is used.
            key_format: The format used to name/index the variable(s).
            keys: If keys: int, it is interpreted as the number of variables to construct.
                  Otherwise, the elements of the sequence are converted to strings via 'str' and
                  substituted into `key_format`.

        Returns:
            A list of variable instances.

        Raises:
            QiskitOptimizationError: if the variable name is already taken.
            QiskitOptimizationError: if less than one variable instantiation is attempted.
            QiskitOptimizationError: if `key_format` has more than one substitution or a
                                     nested substitution.
        """
        return self._var_list(keys, 0, 1, Variable.Type.BINARY, name, key_format)

    def integer_var(
        self,
        lowerbound: Union[float, int] = 0,
        upperbound: Union[float, int] = INFINITY,
        name: Optional[str] = None,
    ) -> Variable:
        """Adds an integer variable to the quadratic program.

        Args:
            lowerbound: The lowerbound of the variable.
            upperbound: The upperbound of the variable.
            name: The name of the variable.
                If it's ``None`` or empty ``""``, the default name, e.g., ``x0``, is used.

        Returns:
            The added variable.

        Raises:
            QiskitOptimizationError: if the variable name is already occupied.
        """
        return self._add_variable(lowerbound, upperbound, Variable.Type.INTEGER, name)

    def integer_var_dict(
        self,
        keys: Union[int, Sequence],
        lowerbound: Union[float, int] = 0,
        upperbound: Union[float, int] = INFINITY,
        name: Optional[str] = None,
        key_format: str = "{}",
    ) -> Dict[str, Variable]:
        """
        Uses 'var_dict' to construct a dictionary of integer variables

        Args:
            lowerbound: The lower bound of the variable(s).
            upperbound: The upper bound of the variable(s).
            name: The name(s) of the variable(s).
                If it's ``None`` or empty ``""``, the default name, e.g., ``x0``, is used.
            key_format: The format used to name/index the variable(s).
            keys: If keys: int, it is interpreted as the number of variables to construct.
                  Otherwise, the elements of the sequence are converted to strings via 'str' and
                  substituted into `key_format`.

        Returns:
            A dictionary mapping the variable names to variable instances.

        Raises:
            QiskitOptimizationError: if the variable name is already taken.
            QiskitOptimizationError: if less than one variable instantiation is attempted.
            QiskitOptimizationError: if `key_format` has more than one substitution or a
                                     nested substitution.
        """
        return self._var_dict(
            keys, lowerbound, upperbound, Variable.Type.INTEGER, name, key_format
        )

    def integer_var_list(
        self,
        keys: Union[int, Sequence],
        lowerbound: Union[float, int] = 0,
        upperbound: Union[float, int] = INFINITY,
        name: Optional[str] = None,
        key_format: str = "{}",
    ) -> List[Variable]:
        """
        Uses 'var_list' to construct a list of integer variables

        Args:
            lowerbound: The lower bound of the variable(s).
            upperbound: The upper bound of the variable(s).
            name: The name(s) of the variable(s).
                If it's ``None`` or empty ``""``, the default name, e.g., ``x0``, is used.
            key_format: The format used to name/index the variable(s).
            keys: If keys: int, it is interpreted as the number of variables to construct.
                  Otherwise, the elements of the sequence are converted to strings via 'str' and
                  substituted into `key_format`.

        Returns:
            A list of variable instances.

        Raises:
            QiskitOptimizationError: if the variable name is already taken.
            QiskitOptimizationError: if less than one variable instantiation is attempted.
            QiskitOptimizationError: if `key_format` has more than one substitution or a
                                     nested substitution.
        """
        return self._var_list(
            keys, lowerbound, upperbound, Variable.Type.INTEGER, name, key_format
        )

    def get_variable(self, i: Union[int, str]) -> Variable:
        """Returns a variable for a given name or index.

        Args:
            i: the index or name of the variable.

        Returns:
            The corresponding variable.
        """
        if isinstance(i, (int, np.integer)):
            return self.variables[i]
        else:
            return self.variables[self._variables_index[i]]

    def get_num_vars(self, vartype: Optional[VarType] = None) -> int:
        """Returns the total number of variables or the number of variables of the specified type.

        Args:
            vartype: The type to be filtered on. All variables are counted if None.

        Returns:
            The total number of variables.
        """
        if vartype:
            return sum(variable.vartype == vartype for variable in self._variables)
        else:
            return len(self._variables)

    def get_num_continuous_vars(self) -> int:
        """Returns the total number of continuous variables.

        Returns:
            The total number of continuous variables.
        """
        return self.get_num_vars(Variable.Type.CONTINUOUS)

    def get_num_binary_vars(self) -> int:
        """Returns the total number of binary variables.

        Returns:
            The total number of binary variables.
        """
        return self.get_num_vars(Variable.Type.BINARY)

    def get_num_integer_vars(self) -> int:
        """Returns the total number of integer variables.

        Returns:
            The total number of integer variables.
        """
        return self.get_num_vars(Variable.Type.INTEGER)

    @property
    def linear_constraints(self) -> List[LinearConstraint]:
        """Returns the list of linear constraints of the quadratic program.

        Returns:
            List of linear constraints.
        """
        return self._linear_constraints

    @property
    def linear_constraints_index(self) -> Dict[str, int]:
        """Returns the dictionary that maps the name of a linear constraint to its index.

        Returns:
            The linear constraint index dictionary.
        """
        return self._linear_constraints_index

    def linear_constraint(
        self,
        linear: Union[
            ndarray, spmatrix, List[float], Dict[Union[int, str], float]
        ] = None,
        sense: Union[str, ConstraintSense] = "<=",
        rhs: float = 0.0,
        name: Optional[str] = None,
    ) -> LinearConstraint:
        """Adds a linear equality constraint to the quadratic program of the form:
            ``(linear * x) sense rhs``.

        Args:
            linear: The linear coefficients of the left-hand side of the constraint.
            sense: The sense of the constraint,

                - ``==``, ``=``, ``E``, and ``EQ`` denote 'equal to'.
                - ``>=``, ``>``, ``G``, and ``GE`` denote 'greater-than-or-equal-to'.
                - ``<=``, ``<``, ``L``, and ``LE`` denote 'less-than-or-equal-to'.

            rhs: The right-hand side of the constraint.
            name: The name of the constraint.
                If it's ``None`` or empty ``""``, the default name, e.g., ``c0``, is used.

        Returns:
            The added constraint.

        Raises:
            QiskitOptimizationError: if the constraint name already exists or the sense is not
                valid.
        """
        if name:
            if name in self.linear_constraints_index:
                raise QiskitOptimizationError(
                    f"Linear constraint's name already exists: {name}"
                )
            self._check_name(name, "Linear constraint")
        else:
            k = self.get_num_linear_constraints()
            while f"c{k}" in self.linear_constraints_index:
                k += 1
            name = f"c{k}"
        self.linear_constraints_index[name] = len(self.linear_constraints)
        if linear is None:
            linear = {}
        constraint = LinearConstraint(
            self, name, linear, Constraint.Sense.convert(sense), rhs
        )
        self.linear_constraints.append(constraint)
        return constraint

    def get_linear_constraint(self, i: Union[int, str]) -> LinearConstraint:
        """Returns a linear constraint for a given name or index.

        Args:
            i: the index or name of the constraint.

        Returns:
            The corresponding constraint.

        Raises:
            IndexError: if the index is out of the list size
            KeyError: if the name does not exist
        """
        if isinstance(i, int):
            return self._linear_constraints[i]
        else:
            return self._linear_constraints[self._linear_constraints_index[i]]

    def get_num_linear_constraints(self) -> int:
        """Returns the number of linear constraints.

        Returns:
            The number of linear constraints.
        """
        return len(self._linear_constraints)

    @property
    def quadratic_constraints(self) -> List[QuadraticConstraint]:
        """Returns the list of quadratic constraints of the quadratic program.

        Returns:
            List of quadratic constraints.
        """
        return self._quadratic_constraints

    @property
    def quadratic_constraints_index(self) -> Dict[str, int]:
        """Returns the dictionary that maps the name of a quadratic constraint to its index.

        Returns:
            The quadratic constraint index dictionary.
        """
        return self._quadratic_constraints_index

    def quadratic_constraint(
        self,
        linear: Union[
            ndarray, spmatrix, List[float], Dict[Union[int, str], float]
        ] = None,
        quadratic: Union[
            ndarray,
            spmatrix,
            List[List[float]],
            Dict[Tuple[Union[int, str], Union[int, str]], float],
        ] = None,
        sense: Union[str, ConstraintSense] = "<=",
        rhs: float = 0.0,
        name: Optional[str] = None,
    ) -> QuadraticConstraint:
        """Adds a quadratic equality constraint to the quadratic program of the form:
            ``(x * quadratic * x + linear * x) sense rhs``.

        Args:
            linear: The linear coefficients of the constraint.
            quadratic: The quadratic coefficients of the constraint.
            sense: The sense of the constraint,

                - ``==``, ``=``, ``E``, and ``EQ`` denote 'equal to'.
                - ``>=``, ``>``, ``G``, and ``GE`` denote 'greater-than-or-equal-to'.
                - ``<=``, ``<``, ``L``, and ``LE`` denote 'less-than-or-equal-to'.

            rhs: The right-hand side of the constraint.
            name: The name of the constraint.
                If it's ``None`` or empty ``""``, the default name, e.g., ``q0``, is used.

        Returns:
            The added constraint.

        Raises:
            QiskitOptimizationError: if the constraint name already exists.
        """
        if name:
            if name in self.quadratic_constraints_index:
                raise QiskitOptimizationError(
                    f"Quadratic constraint name already exists: {name}"
                )
            self._check_name(name, "Quadratic constraint")
        else:
            k = self.get_num_quadratic_constraints()
            while f"q{k}" in self.quadratic_constraints_index:
                k += 1
            name = f"q{k}"
        self.quadratic_constraints_index[name] = len(self.quadratic_constraints)
        if linear is None:
            linear = {}
        if quadratic is None:
            quadratic = {}
        constraint = QuadraticConstraint(
            self, name, linear, quadratic, Constraint.Sense.convert(sense), rhs
        )
        self.quadratic_constraints.append(constraint)
        return constraint

    def get_quadratic_constraint(self, i: Union[int, str]) -> QuadraticConstraint:
        """Returns a quadratic constraint for a given name or index.

        Args:
            i: the index or name of the constraint.

        Returns:
            The corresponding constraint.

        Raises:
            IndexError: if the index is out of the list size
            KeyError: if the name does not exist
        """
        if isinstance(i, int):
            return self._quadratic_constraints[i]
        else:
            return self._quadratic_constraints[self._quadratic_constraints_index[i]]

    def get_num_quadratic_constraints(self) -> int:
        """Returns the number of quadratic constraints.

        Returns:
            The number of quadratic constraints.
        """
        return len(self._quadratic_constraints)

    def remove_linear_constraint(self, i: Union[str, int]) -> None:
        """Remove a linear constraint

        Args:
            i: an index or a name of a linear constraint

        Raises:
            KeyError: if name does not exist
            IndexError: if index is out of range
        """
        if isinstance(i, str):
            i = self._linear_constraints_index[i]
        del self._linear_constraints[i]
        self._linear_constraints_index = {
            cst.name: j for j, cst in enumerate(self._linear_constraints)
        }

    def remove_quadratic_constraint(self, i: Union[str, int]) -> None:
        """Remove a quadratic constraint

        Args:
            i: an index or a name of a quadratic constraint

        Raises:
            KeyError: if name does not exist
            IndexError: if index is out of range
        """
        if isinstance(i, str):
            i = self._quadratic_constraints_index[i]
        del self._quadratic_constraints[i]
        self._quadratic_constraints_index = {
            cst.name: j for j, cst in enumerate(self._quadratic_constraints)
        }

    @property
    def objective(self) -> QuadraticObjective:
        """Returns the quadratic objective.

        Returns:
            The quadratic objective.
        """
        return self._objective

    def minimize(
        self,
        constant: float = 0.0,
        linear: Union[
            ndarray, spmatrix, List[float], Dict[Union[str, int], float]
        ] = None,
        quadratic: Union[
            ndarray,
            spmatrix,
            List[List[float]],
            Dict[Tuple[Union[int, str], Union[int, str]], float],
        ] = None,
    ) -> None:
        """Sets a quadratic objective to be minimized.

        Args:
            constant: the constant offset of the objective.
            linear: the coefficients of the linear part of the objective.
            quadratic: the coefficients of the quadratic part of the objective.

        Returns:
            The created quadratic objective.
        """
        self._objective = QuadraticObjective(
            self, constant, linear, quadratic, QuadraticObjective.Sense.MINIMIZE
        )

    def maximize(
        self,
        constant: float = 0.0,
        linear: Union[
            ndarray, spmatrix, List[float], Dict[Union[str, int], float]
        ] = None,
        quadratic: Union[
            ndarray,
            spmatrix,
            List[List[float]],
            Dict[Tuple[Union[int, str], Union[int, str]], float],
        ] = None,
    ) -> None:
        """Sets a quadratic objective to be maximized.

        Args:
            constant: the constant offset of the objective.
            linear: the coefficients of the linear part of the objective.
            quadratic: the coefficients of the quadratic part of the objective.

        Returns:
            The created quadratic objective.
        """
        self._objective = QuadraticObjective(
            self, constant, linear, quadratic, QuadraticObjective.Sense.MAXIMIZE
        )

    def _copy_from(self, other: "QuadraticProgram", include_name: bool) -> None:
        """Copy another QuadraticProgram to this updating QuadraticProgramElement

        Note: this breaks the consistency of `other`. You cannot use `other` after the copy.

        Args:
            other: The quadratic program to be copied from.
            include_name: Whether this method copies the problem name or not.
        """
        for attr, val in vars(other).items():
            if attr == "_name" and not include_name:
                continue
            if isinstance(val, QuadraticProgramElement):
                val.quadratic_program = self
            if isinstance(val, list):
                for elem in val:
                    if isinstance(elem, QuadraticProgramElement):
                        elem.quadratic_program = self
            setattr(self, attr, val)

    def substitute_variables(
        self,
        constants: Optional[Dict[Union[str, int], float]] = None,
        variables: Optional[
            Dict[Union[str, int], Tuple[Union[str, int], float]]
        ] = None,
    ) -> "QuadraticProgram":
        """Substitutes variables with constants or other variables.

        Args:
            constants: replace variable by constant
                e.g., ``{'x': 2}`` means ``x`` is substituted with 2

            variables: replace variables by weighted other variable
                need to copy everything using name reference to make sure that indices are matched
                correctly. The lower and upper bounds are updated accordingly.
                e.g., ``{'x': ('y', 2)}`` means ``x`` is substituted with ``y * 2``

        Returns:
            An optimization problem by substituting variables with constants or other variables.
            If the substitution is valid, ``QuadraticProgram.status`` is still
            ``QuadraticProgram.Status.VALID``.
            Otherwise, it gets ``QuadraticProgram.Status.INFEASIBLE``.

        Raises:
            QiskitOptimizationError: if the substitution is invalid as follows.

                - Same variable is substituted multiple times.
                - Coefficient of variable substitution is zero.
        """
        # pylint: disable=cyclic-import
        from .substitute_variables import substitute_variables

        return substitute_variables(self, constants, variables)

    def to_ising(self) -> Tuple[SparsePauliOp, float]:
        """Return the Ising Hamiltonian of this problem.

        Variables are mapped to qubits in the same order, i.e.,
        i-th variable is mapped to i-th qubit.
        See https://github.com/Qiskit/qiskit-terra/issues/1148 for details.

        Returns:
            qubit_op: The qubit operator for the problem
            offset: The constant value in the Ising Hamiltonian.

        Raises:
            QiskitOptimizationError: If a variable type is not binary.
            QiskitOptimizationError: If constraints exist in the problem.
        """
        # pylint: disable=cyclic-import
        from ..translators.ising import to_ising

        return to_ising(self)

    def from_ising(
        self,
        qubit_op: BaseOperator,
        offset: float = 0.0,
        linear: bool = False,
    ) -> None:
        r"""Create a quadratic program from a qubit operator and a shift value.

        Variables are mapped to qubits in the same order, i.e.,
        i-th variable is mapped to i-th qubit.
        See https://github.com/Qiskit/qiskit-terra/issues/1148 for details.

        Args:
            qubit_op: The qubit operator of the problem.
            offset: The constant value in the Ising Hamiltonian.
            linear: If linear is True, :math:`x^2` is treated as a linear term
                since :math:`x^2 = x` for :math:`x \in \{0,1\}`.
                Else, :math:`x^2` is treated as a quadratic term.
                The default value is False.

        Raises:
            QiskitOptimizationError: If there are Pauli Xs in any Pauli term
            QiskitOptimizationError: If there are more than 2 Pauli Zs in any Pauli term
            NotImplementedError: If the input operator is a ListOp
        """
        # pylint: disable=cyclic-import
        from ..translators.ising import from_ising

        other = from_ising(qubit_op, offset, linear)
        self._copy_from(other, include_name=False)

    def get_feasibility_info(
        self, x: Union[List[float], np.ndarray]
    ) -> Tuple[bool, List[Variable], List[Constraint]]:
        """Returns whether a solution is feasible or not along with the violations.
        Args:
            x: a solution value, such as returned in an optimizer result.
        Returns:
            feasible: Whether the solution provided is feasible or not.
            List[Variable]: List of variables which are violated.
            List[Constraint]: List of constraints which are violated.

        Raises:
            QiskitOptimizationError: If the input `x` is not same len as total vars
        """
        # if input `x` is not the same len as the total vars, raise an error
        if len(x) != self.get_num_vars():
            raise QiskitOptimizationError(
                f"The size of solution `x`: {len(x)}, does not match the number of problem variables: "
                f"{self.get_num_vars()}"
            )

        # check whether the input satisfy the bounds of the problem
        violated_variables = []
        for i, val in enumerate(x):
            variable = self.get_variable(i)
            if val < variable.lowerbound or variable.upperbound < val:
                violated_variables.append(variable)

        # check whether the input satisfy the constraints of the problem
        violated_constraints = []
        for constraint in cast(List[Constraint], self._linear_constraints) + cast(
            List[Constraint], self._quadratic_constraints
        ):
            lhs = constraint.evaluate(x)
            if constraint.sense == ConstraintSense.LE and lhs > constraint.rhs:
                violated_constraints.append(constraint)
            elif constraint.sense == ConstraintSense.GE and lhs < constraint.rhs:
                violated_constraints.append(constraint)
            elif constraint.sense == ConstraintSense.EQ and not isclose(
                lhs, constraint.rhs
            ):
                violated_constraints.append(constraint)

        feasible = not violated_variables and not violated_constraints

        return feasible, violated_variables, violated_constraints

    def is_feasible(self, x: Union[List[float], np.ndarray]) -> bool:
        """Returns whether a solution is feasible or not.

        Args:
            x: a solution value, such as returned in an optimizer result.

        Returns:
            ``True`` if the solution provided is feasible otherwise ``False``.

        """
        feasible, _, _ = self.get_feasibility_info(x)

        return feasible

    def prettyprint(self, wrap: int = 80) -> str:
        """Returns a pretty printed string of this problem.

        Args:
            wrap: The text width to wrap the output strings. It is disabled by setting 0.
                Note that some strings might exceed this value, for example, a long variable
                name won't be wrapped. The default value is 80.

        Returns:
            A pretty printed string representing the problem.

        Raises:
            QiskitOptimizationError: if there is a non-printable name.
        """
        # pylint: disable=cyclic-import
        from ..translators.prettyprint import prettyprint

        return prettyprint(self, wrap)

    @staticmethod
    def _check_name(name: str, name_type: str) -> None:
        """Displays a warning message if a name string is not printable"""
        if not name.isprintable():
            warn(f"{name_type} name is not printable: {repr(name)}")



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Quadratic expression interface."""

from typing import List, Union, Dict, Tuple, Any

import numpy as np
from numpy import ndarray
from scipy.sparse import spmatrix, dok_matrix, tril, triu

from .quadratic_program_element import QuadraticProgramElement
from ..exceptions import QiskitOptimizationError
from ..infinity import INFINITY
from .linear_expression import ExpressionBounds


class QuadraticExpression(QuadraticProgramElement):
    """Representation of a quadratic expression by its coefficients."""

    def __init__(
        self,
        quadratic_program: Any,
        coefficients: Union[
            ndarray,
            spmatrix,
            List[List[float]],
            Dict[Tuple[Union[int, str], Union[int, str]], float],
        ],
    ) -> None:
        """Creates a new quadratic expression.

        The quadratic expression can be defined via an array, a list, a sparse matrix, or a
        dictionary that uses variable names or indices as keys and stores the values internally as a
        dok_matrix. We stores values in a compressed way, i.e., values at symmetric positions are
        summed up in the upper triangle. For example, {(0, 1): 1, (1, 0): 2} -> {(0, 1): 3}.

        Args:
            quadratic_program: The parent QuadraticProgram.
            coefficients: The (sparse) representation of the coefficients.

        """
        super().__init__(quadratic_program)
        self.coefficients = coefficients

    def __getitem__(self, key: Tuple[Union[int, str], Union[int, str]]) -> float:
        """Returns the coefficient where i, j can be a variable names or indices.

        Args:
            key: The tuple of indices or names of the variables corresponding to the coefficient.

        Returns:
            The coefficient corresponding to the addressed variables.
        """
        i, j = key
        if isinstance(i, str):
            i = self.quadratic_program.variables_index[i]
        if isinstance(j, str):
            j = self.quadratic_program.variables_index[j]
        return self.coefficients[min(i, j), max(i, j)]

    def __setitem__(
        self, key: Tuple[Union[int, str], Union[int, str]], value: float
    ) -> None:
        """Sets the coefficient where i, j can be a variable names or indices.

        Args:
            key: The tuple of indices or names of the variables corresponding to the coefficient.
            value: The coefficient corresponding to the addressed variables.
        """
        i, j = key
        if isinstance(i, str):
            i = self.quadratic_program.variables_index[i]
        if isinstance(j, str):
            j = self.quadratic_program.variables_index[j]
        self.coefficients[min(i, j), max(i, j)] = value

    def _coeffs_to_dok_matrix(
        self,
        coefficients: Union[
            ndarray,
            spmatrix,
            List[List[float]],
            Dict[Tuple[Union[int, str], Union[int, str]], float],
        ],
    ) -> dok_matrix:
        """Maps given coefficients to a dok_matrix.

        Args:
            coefficients: The coefficients to be mapped.

        Returns:
            The given coefficients as a dok_matrix

        Raises:
            QiskitOptimizationError: if coefficients are given in unsupported format.
        """
        if isinstance(coefficients, (list, ndarray, spmatrix)):
            coefficients = dok_matrix(coefficients)
        elif isinstance(coefficients, dict):
            n = self.quadratic_program.get_num_vars()
            coeffs = dok_matrix((n, n))
            for (i, j), value in coefficients.items():
                if isinstance(i, str):
                    i = self.quadratic_program.variables_index[i]
                if isinstance(j, str):
                    j = self.quadratic_program.variables_index[j]
                coeffs[i, j] = value
            coefficients = coeffs
        else:
            raise QiskitOptimizationError(
                f"Unsupported format for coefficients: {coefficients}"
            )
        return self._triangle_matrix(coefficients)

    @staticmethod
    def _triangle_matrix(mat: dok_matrix) -> dok_matrix:
        lower = tril(mat, -1, format="dok")
        # `todok` is necessary because subtraction results in other format
        return (mat + lower.transpose() - lower).todok()

    @staticmethod
    def _symmetric_matrix(mat: dok_matrix) -> dok_matrix:
        upper = triu(mat, 1, format="dok") / 2
        # `todok` is necessary because subtraction results in other format
        return (mat + upper.transpose() - upper).todok()

    @property
    def coefficients(self) -> dok_matrix:
        """Returns the coefficients of the quadratic expression.

        Returns:
            The coefficients of the quadratic expression.
        """
        return self._coefficients

    @coefficients.setter
    def coefficients(
        self,
        coefficients: Union[
            ndarray,
            spmatrix,
            List[List[float]],
            Dict[Tuple[Union[int, str], Union[int, str]], float],
        ],
    ) -> None:
        """Sets the coefficients of the quadratic expression.

        Args:
            coefficients: The coefficients of the quadratic expression.
        """
        self._coefficients = self._coeffs_to_dok_matrix(coefficients)

    def to_array(self, symmetric: bool = False) -> ndarray:
        """Returns the coefficients of the quadratic expression as array.

        Args:
            symmetric: Determines whether the output is in a symmetric form or not.

        Returns:
            An array with the coefficients corresponding to the quadratic expression.
        """
        coeffs = (
            self._symmetric_matrix(self._coefficients)
            if symmetric
            else self._coefficients
        )
        return coeffs.toarray()

    def to_dict(
        self, symmetric: bool = False, use_name: bool = False
    ) -> Dict[Union[Tuple[int, int], Tuple[str, str]], float]:
        """Returns the coefficients of the quadratic expression as dictionary, either using tuples
        of variable names or indices as keys.

        Args:
            symmetric: Determines whether the output is in a symmetric form or not.
            use_name: Determines whether to use index or names to refer to variables.

        Returns:
            An dictionary with the coefficients corresponding to the quadratic expression.
        """
        coeffs = (
            self._symmetric_matrix(self._coefficients)
            if symmetric
            else self._coefficients
        )
        if use_name:
            return {
                (
                    self.quadratic_program.variables[i].name,
                    self.quadratic_program.variables[j].name,
                ): v
                for (i, j), v in coeffs.items()
            }
        else:
            return {(int(i), int(j)): v for (i, j), v in coeffs.items()}

    def evaluate(self, x: Union[ndarray, List, Dict[Union[int, str], float]]) -> float:
        """Evaluate the quadratic expression for given variables: x * Q * x.

        Args:
            x: The values of the variables to be evaluated.

        Returns:
            The value of the quadratic expression given the variable values.
        """
        x = self._cast_as_array(x)

        # compute x * Q * x for the quadratic expression
        val = x @ self.coefficients @ x

        # return the result
        return val

    def evaluate_gradient(
        self, x: Union[ndarray, List, Dict[Union[int, str], float]]
    ) -> ndarray:
        """Evaluate the gradient of the quadratic expression for given variables.

        Args:
            x: The values of the variables to be evaluated.

        Returns:
            The value of the gradient quadratic expression given the variable values.
        """
        x = self._cast_as_array(x)

        # compute (Q' + Q) * x for the quadratic expression
        val = (self.coefficients.transpose() + self.coefficients) @ x

        # return the result
        return val

    def _cast_as_array(
        self, x: Union[ndarray, List, Dict[Union[int, str], float]]
    ) -> Union[dok_matrix, np.ndarray]:
        """Converts input to an array if it is a dictionary or list."""
        if isinstance(x, dict):
            x_aux = np.zeros(self.quadratic_program.get_num_vars())
            for i, v in x.items():
                if isinstance(i, str):
                    i = self.quadratic_program.variables_index[i]
                x_aux[i] = v
            x = x_aux
        if isinstance(x, list):
            x = np.array(x)
        return x

    @property
    def bounds(self) -> ExpressionBounds:
        """Returns the lower bound and the upper bound of the quadratic expression

        Returns:
            The lower bound and the upper bound of the quadratic expression

        Raises:
            QiskitOptimizationError: if the quadratic expression contains any unbounded variable
        """
        l_b = u_b = 0.0
        for (ind1, ind2), coeff in self.to_dict().items():
            x = self.quadratic_program.get_variable(ind1)
            if x.lowerbound == -INFINITY or x.upperbound == INFINITY:
                raise QiskitOptimizationError(
                    f"Quadratic expression contains an unbounded variable: {x.name}"
                )
            y = self.quadratic_program.get_variable(ind2)
            if y.lowerbound == -INFINITY or y.upperbound == INFINITY:
                raise QiskitOptimizationError(
                    f"Quadratic expression contains an unbounded variable: {y.name}"
                )
            lst = []
            if ind1 == ind2:
                if x.lowerbound * x.upperbound <= 0.0:
                    # lower bound and upper bound have different signs
                    lst.append(0.0)
                lst.extend([x.lowerbound**2, x.upperbound**2])
            else:
                lst.extend(
                    [
                        x.lowerbound * y.lowerbound,
                        x.lowerbound * y.upperbound,
                        x.upperbound * y.lowerbound,
                        x.upperbound * y.upperbound,
                    ]
                )
            lst2 = [coeff * val for val in lst]
            l_b += min(lst2)
            u_b += max(lst2)
        return ExpressionBounds(lowerbound=l_b, upperbound=u_b)

    def __repr__(self):
        # pylint: disable=cyclic-import
        from ..translators.prettyprint import expr2str, DEFAULT_TRUNCATE

        return f"<{self.__class__.__name__}: {expr2str(quadratic=self, truncate=DEFAULT_TRUNCATE)}>"

    def __str__(self):
        # pylint: disable=cyclic-import
        from ..translators.prettyprint import expr2str

        return f"{expr2str(quadratic=self)}"



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Linear Constraint."""

from typing import Union, List, Dict, Any

from numpy import ndarray
from scipy.sparse import spmatrix

from .constraint import Constraint, ConstraintSense
from .linear_expression import LinearExpression


class LinearConstraint(Constraint):
    """Representation of a linear constraint."""

    # Note: added, duplicating in effect that in Constraint, to avoid issues with Sphinx
    Sense = ConstraintSense

    def __init__(
        self,
        quadratic_program: Any,
        name: str,
        linear: Union[ndarray, spmatrix, List[float], Dict[Union[str, int], float]],
        sense: ConstraintSense,
        rhs: float,
    ) -> None:
        """
        Args:
            quadratic_program: The parent quadratic program.
            name: The name of the constraint.
            linear: The coefficients specifying the linear constraint.
            sense: The sense of the constraint.
            rhs: The right-hand-side of the constraint.
        """
        super().__init__(quadratic_program, name, sense, rhs)
        self._linear = LinearExpression(quadratic_program, linear)

    @property
    def linear(self) -> LinearExpression:
        """Returns the linear expression corresponding to the left-hand-side of the constraint.

        Returns:
            The left-hand-side linear expression.
        """
        return self._linear

    @linear.setter
    def linear(
        self,
        linear: Union[ndarray, spmatrix, List[float], Dict[Union[str, int], float]],
    ) -> None:
        """Sets the linear expression corresponding to the left-hand-side of the constraint.
        The coefficients can either be given by an array, a (sparse) 1d matrix, a list or a
        dictionary.

        Args:
            linear: The linear coefficients of the left-hand-side.
        """
        self._linear = LinearExpression(self.quadratic_program, linear)

    def evaluate(self, x: Union[ndarray, List, Dict[Union[int, str], float]]) -> float:
        """Evaluate the left-hand-side of the constraint.

        Args:
            x: The values of the variables to be evaluated.

        Returns:
            The left-hand-side of the constraint given the variable values.
        """
        return self.linear.evaluate(x)

    def __repr__(self):
        # pylint: disable=cyclic-import
        from ..translators.prettyprint import expr2str, DEFAULT_TRUNCATE

        lhs = expr2str(linear=self.linear, truncate=DEFAULT_TRUNCATE)
        return f"<{self.__class__.__name__}: {lhs} {self.sense.label} {self.rhs} '{self.name}'>"

    def __str__(self):
        # pylint: disable=cyclic-import
        from ..translators.prettyprint import expr2str

        lhs = expr2str(linear=self.linear)
        return f"{lhs} {self.sense.label} {self.rhs} '{self.name}'"



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2020, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""The inequality to equality converter."""

import copy
import math
from typing import List, Optional, Union

import numpy as np

from ..exceptions import QiskitOptimizationError
from ..problems.constraint import Constraint
from ..problems.linear_constraint import LinearConstraint
from ..problems.quadratic_constraint import QuadraticConstraint
from ..problems.quadratic_objective import QuadraticObjective
from ..problems.quadratic_program import QuadraticProgram
from ..problems.variable import Variable
from .quadratic_program_converter import QuadraticProgramConverter


class InequalityToEquality(QuadraticProgramConverter):
    """Convert inequality constraints into equality constraints by introducing slack variables.

    Examples:
        >>> from qiskit_optimization.problems import QuadraticProgram
        >>> from qiskit_optimization.converters import InequalityToEquality
        >>> problem = QuadraticProgram()
        >>> # define a problem
        >>> conv = InequalityToEquality()
        >>> problem2 = conv.convert(problem)
    """

    _delimiter = "@"  # users are supposed not to use this character in variable names

    def __init__(self, mode: str = "auto") -> None:
        """
        Args:
            mode: To choose the type of slack variables. There are 3 options for mode.

                - 'integer': All slack variables will be integer variables.
                - 'continuous': All slack variables will be continuous variables.
                - 'auto': Use integer variables if possible, otherwise use continuous variables.
        """
        self._src: Optional[QuadraticProgram] = None
        self._dst: Optional[QuadraticProgram] = None
        self._mode = mode

    def convert(self, problem: QuadraticProgram) -> QuadraticProgram:
        """Convert a problem with inequality constraints into one with only equality constraints.

        Args:
            problem: The problem to be solved, that may contain inequality constraints.

        Returns:
            The converted problem, that contain only equality constraints.

        Raises:
            QiskitOptimizationError: If a variable type is not supported.
            QiskitOptimizationError: If an unsupported mode is selected.
            QiskitOptimizationError: If an unsupported sense is specified.
        """
        self._src = copy.deepcopy(problem)
        self._dst = QuadraticProgram(name=problem.name)

        # set a converting mode
        mode = self._mode
        if mode not in ["integer", "continuous", "auto"]:
            raise QiskitOptimizationError(f"Unsupported mode is selected: {mode}")

        # Copy variables
        for x in self._src.variables:
            if x.vartype == Variable.Type.BINARY:
                self._dst.binary_var(name=x.name)
            elif x.vartype == Variable.Type.INTEGER:
                self._dst.integer_var(
                    name=x.name, lowerbound=x.lowerbound, upperbound=x.upperbound
                )
            elif x.vartype == Variable.Type.CONTINUOUS:
                self._dst.continuous_var(
                    name=x.name, lowerbound=x.lowerbound, upperbound=x.upperbound
                )
            else:
                raise QiskitOptimizationError(f"Unsupported variable type {x.vartype}")

        # Note: QuadraticProgram needs to add all variables before adding any constraints.

        # Add slack variables to linear constraints
        new_linear_constraints = []
        for lin_const in self._src.linear_constraints:
            if lin_const.sense == Constraint.Sense.EQ:
                new_linear_constraints.append(
                    (
                        lin_const.linear.coefficients,
                        lin_const.sense,
                        lin_const.rhs,
                        lin_const.name,
                    )
                )
            elif lin_const.sense in [Constraint.Sense.LE, Constraint.Sense.GE]:
                new_linear_constraints.append(
                    self._add_slack_var_linear_constraint(lin_const)
                )
            else:
                raise QiskitOptimizationError(
                    f"Internal error: type of sense in {lin_const.name} is not supported: "
                    f"{lin_const.sense}"
                )

        # Add slack variables to quadratic constraints
        new_quadratic_constraints = []
        for quad_const in self._src.quadratic_constraints:
            if quad_const.sense == Constraint.Sense.EQ:
                new_quadratic_constraints.append(
                    (
                        quad_const.linear.coefficients,
                        quad_const.quadratic.coefficients,
                        quad_const.sense,
                        quad_const.rhs,
                        quad_const.name,
                    )
                )
            elif quad_const.sense in [Constraint.Sense.LE, Constraint.Sense.GE]:
                new_quadratic_constraints.append(
                    self._add_slack_var_quadratic_constraint(quad_const)
                )
            else:
                raise QiskitOptimizationError(
                    f"Internal error: type of sense in {quad_const.name} is not supported: "
                    f"{quad_const.sense}"
                )

        # Copy the objective function
        constant = self._src.objective.constant
        linear = self._src.objective.linear.to_dict(use_name=True)
        quadratic = self._src.objective.quadratic.to_dict(use_name=True)
        if self._src.objective.sense == QuadraticObjective.Sense.MINIMIZE:
            self._dst.minimize(constant, linear, quadratic)
        else:
            self._dst.maximize(constant, linear, quadratic)

        # Add linear constraints
        for lin_const_args in new_linear_constraints:
            self._dst.linear_constraint(*lin_const_args)

        # Add quadratic constraints
        for quad_const_args in new_quadratic_constraints:
            self._dst.quadratic_constraint(*quad_const_args)

        return self._dst

    def _add_slack_var_linear_constraint(self, constraint: LinearConstraint):
        linear = constraint.linear
        sense = constraint.sense
        name = constraint.name

        any_float = self._any_float(linear.to_array())
        mode = self._mode
        if mode == "integer":
            if any_float:
                raise QiskitOptimizationError(
                    f'"{name}" contains float coefficients. '
                    'We can not use an integer slack variable for "{name}"'
                )
        elif mode == "auto":
            mode = "continuous" if any_float else "integer"

        new_rhs = constraint.rhs
        if mode == "integer":
            # If rhs is float number, round up/down to the nearest integer.
            if sense == Constraint.Sense.LE:
                new_rhs = math.floor(new_rhs)
            if sense == Constraint.Sense.GE:
                new_rhs = math.ceil(new_rhs)

        lin_bounds = linear.bounds
        lhs_lb = lin_bounds.lowerbound
        lhs_ub = lin_bounds.upperbound

        var_ub = 0.0
        sign = 0
        if sense == Constraint.Sense.LE:
            var_ub = new_rhs - lhs_lb
            if var_ub > 0:
                sign = 1
        elif sense == Constraint.Sense.GE:
            var_ub = lhs_ub - new_rhs
            if var_ub > 0:
                sign = -1

        new_linear = linear.to_dict(use_name=True)
        if var_ub > 0:
            # Add a slack variable.
            mode_name = {"integer": "int", "continuous": "continuous"}
            slack_name = f"{name}{self._delimiter}{mode_name[mode]}_slack"
            if mode == "integer":
                self._dst.integer_var(name=slack_name, lowerbound=0, upperbound=var_ub)
            elif mode == "continuous":
                self._dst.continuous_var(
                    name=slack_name, lowerbound=0, upperbound=var_ub
                )
            new_linear[slack_name] = sign
        return new_linear, "==", new_rhs, name

    def _add_slack_var_quadratic_constraint(self, constraint: QuadraticConstraint):
        quadratic = constraint.quadratic
        linear = constraint.linear
        sense = constraint.sense
        name = constraint.name

        any_float = self._any_float(linear.to_array()) or self._any_float(
            quadratic.to_array()
        )
        mode = self._mode
        if mode == "integer":
            if any_float:
                raise QiskitOptimizationError(
                    f'"{name}" contains float coefficients. '
                    'We can not use an integer slack variable for "{name}"'
                )
        elif mode == "auto":
            mode = "continuous" if any_float else "integer"

        new_rhs = constraint.rhs
        if mode == "integer":
            # If rhs is float number, round up/down to the nearest integer.
            if sense == Constraint.Sense.LE:
                new_rhs = math.floor(new_rhs)
            if sense == Constraint.Sense.GE:
                new_rhs = math.ceil(new_rhs)

        lin_bounds = linear.bounds
        quad_bounds = quadratic.bounds
        lhs_lb = lin_bounds.lowerbound + quad_bounds.lowerbound
        lhs_ub = lin_bounds.upperbound + quad_bounds.upperbound

        var_ub = 0.0
        sign = 0
        if sense == Constraint.Sense.LE:
            var_ub = new_rhs - lhs_lb
            if var_ub > 0:
                sign = 1
        elif sense == Constraint.Sense.GE:
            var_ub = lhs_ub - new_rhs
            if var_ub > 0:
                sign = -1

        new_linear = linear.to_dict(use_name=True)
        if var_ub > 0:
            # Add a slack variable.
            mode_name = {"integer": "int", "continuous": "continuous"}
            slack_name = f"{name}{self._delimiter}{mode_name[mode]}_slack"
            if mode == "integer":
                self._dst.integer_var(name=slack_name, lowerbound=0, upperbound=var_ub)
            elif mode == "continuous":
                self._dst.continuous_var(
                    name=slack_name, lowerbound=0, upperbound=var_ub
                )
            new_linear[slack_name] = sign
        return new_linear, quadratic.coefficients, "==", new_rhs, name

    def interpret(self, x: Union[np.ndarray, List[float]]) -> np.ndarray:
        """Convert a result of a converted problem into that of the original problem.

        Args:
            x: The result of the converted problem or the given result in case of FAILURE.

        Returns:
            The result of the original problem.
        """
        # convert back the optimization result into that of the original problem
        names = [var.name for var in self._dst.variables]

        # interpret slack variables
        sol = {name: x[i] for i, name in enumerate(names)}
        new_x = np.zeros(self._src.get_num_vars())
        for i, var in enumerate(self._src.variables):
            new_x[i] = sol[var.name]
        return new_x

    @staticmethod
    def _any_float(values: np.ndarray) -> bool:
        """Check whether the list contains float or not.
        This method is used to check whether a constraint contain float coefficients or not.

        Args:
            values: Coefficients of the constraint

        Returns:
            bool: If the constraint contains float coefficients, this returns True, else False.
        """
        return any(isinstance(v, float) and not v.is_integer() for v in values)

    @property
    def mode(self) -> str:
        """Returns the mode of the converter

        Returns:
            The mode of the converter used for additional slack variables
        """
        return self._mode

    @mode.setter
    def mode(self, mode: str) -> None:
        """Set a new mode for the converter

        Args:
            mode: The new mode for the converter
        """
        self._mode = mode



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union, TypeVar
from enum import Enum
import datetime
from uuid import UUID

from pydantic import BaseModel

from q3as.encoding import EncodedVQE, EncodedVQEResult
from q3as.run_options import RunOptions

if TYPE_CHECKING:
    from q3as.client import Client
    from q3as.algo.vqe import VQE, VQEResult

T = TypeVar("T")


class JobRequest(BaseModel):
    """
    Request to run a VQE job.
    """

    input: EncodedVQE
    "The input VQE to run."
    run_options: RunOptions = RunOptions()
    "The options to run the VQE with."

    def send(self, client: Client) -> Job:
        """
        Send the request to the server and create a job.
        """
        return client.create_job(self)


class JobStatus(Enum):
    """
    Enum representing the status of a job.
    """

    STARTED = 0
    "The job is started."
    PAUSED = 1
    "The job is paused."
    SUCCESS = 2
    "The job finished successfully."
    ERROR = 3
    "The job finished with an error."


class JobInfo(BaseModel):
    id: UUID
    slug: str
    request: JobRequest
    status: JobStatus
    result: Optional[Union[EncodedVQEResult, str]]
    created_at: datetime.datetime
    updated_at: Optional[datetime.datetime]
    finished_at: Optional[datetime.datetime]


class BaseJob:
    """
    Base class for a job.
    """

    info: JobInfo

    def __init__(self, info: JobInfo):
        self.info = info

    @property
    def name(self) -> str:
        """
        The name of the job.
        """
        return self.info.slug

    @property
    def status(self) -> JobStatus:
        """
        The status of the job.
        """
        return self.info.status

    @property
    def input(self) -> VQE:
        """
        The input VQE.
        """
        return self.info.request.input.decode()

    @property
    def run_options(self) -> RunOptions:
        """
        The run options.
        """
        return self.info.request.run_options

    @property
    def created_at(self) -> datetime.datetime:
        """
        The time the job was created.
        """
        return self.info.created_at

    @property
    def updated_at(self) -> Optional[datetime.datetime]:
        """
        The time the job was last updated.
        """
        return self.info.updated_at

    @property
    def finished_at(self) -> Optional[datetime.datetime]:
        """
        The time the job finished.
        """
        return self.info.finished_at

    def result_now(self) -> Optional[VQEResult]:
        """
        Get the result of the job without waiting for it to finish. May return None if the job is not finished, or raise an exception if the job failed.
        """
        if self.info.result is None:
            return None
        if isinstance(self.info.result, str):
            raise Exception(self.info.result)
        return self.info.result.decode()


class Job(BaseJob):
    client: Client

    def __init__(self, client: Client, info: JobInfo):
        super().__init__(info)
        self.client = client

    def wait(
        self, polling_interval: float = 1.0, max_wait: Optional[float] = None
    ) -> Job:
        """
        Wait for the job to finish
        """
        if self.status is not JobStatus.STARTED:
            return self
        job = self.client.wait_for_job(self.info.slug, polling_interval, max_wait)
        self.info = job.info
        return job

    def result(
        self, polling_interval: float = 1.0, max_wait: Optional[float] = None
    ) -> VQEResult:
        """
        Get the result of the job, waiting for it to finish if necessary.
        """
        job = self.wait(polling_interval, max_wait)
        result_now = job.result_now()
        if result_now is None:
            raise ValueError("Job did not return a result")
        return result_now

    def refetch(self) -> Job:
        """
        Refetch the job from the server.
        """
        job = self.client.get_job(self.info.slug)
        self.info = job.info
        return job

    def pause(self) -> Job:
        """
        Pause the job.
        """
        job = self.client.pause_job(self.info.slug)
        self.info = job.info
        return job

    def resume(self) -> Job:
        """
        Resume the job.
        """
        job = self.client.resume_job(self.info.slug)
        self.info = job.info
        return job

    def delete(self) -> Job:
        """
        Delete the job.
        """
        job = self.client.delete_job(self.info.slug)
        self.info = job.info
        return job



from __future__ import annotations
from typing import TYPE_CHECKING, cast, Union, Dict, Tuple, List

from pydantic import BaseModel

import q3as.quadratic.problems as qprob
from q3as.quadratic.problems.types import VarType, QuadraticProgramStatus, ObjSense

if TYPE_CHECKING:
    from q3as.quadratic.problems import (
        QuadraticProgram,
        Variable,
        LinearConstraint,
        QuadraticConstraint,
        QuadraticObjective,
        LinearExpression,
        QuadraticExpression,
    )


class EncodedVariable(BaseModel):
    name: str
    lowerbound: Union[float, int]
    upperbound: Union[float, int]
    vartype: VarType

    @classmethod
    def encode(cls, var: Variable) -> EncodedVariable:
        return cls(
            name=var.name,
            lowerbound=var.lowerbound,
            upperbound=var.upperbound,
            vartype=var.vartype,
        )

    def decode(self, qp: QuadraticProgram) -> Variable:
        return qprob.Variable(
            qp, self.name, self.lowerbound, self.upperbound, self.vartype
        )


type EncodedLinearExpression = Dict[int, float]


def encode_linear_expression(exp: LinearExpression) -> EncodedLinearExpression:
    return cast(EncodedLinearExpression, exp.to_dict())


def decode_linear_expression(
    exp: EncodedLinearExpression,
) -> Dict[Union[int, str], float]:
    return cast(Dict[Union[int, str], float], exp)


class EncodedLinearConstraint(BaseModel):
    name: str
    linear: EncodedLinearExpression
    sense: str
    rhs: float

    @classmethod
    def encode(cls, constraint: LinearConstraint) -> EncodedLinearConstraint:
        return cls(
            name=constraint.name,
            linear=encode_linear_expression(constraint.linear),
            sense=constraint.sense.name,
            rhs=constraint.rhs,
        )

    def decode(self, qp: QuadraticProgram) -> LinearConstraint:
        return qprob.LinearConstraint(
            qp,
            self.name,
            decode_linear_expression(self.linear),
            qprob.constraint.ConstraintSense.convert(self.sense),
            self.rhs,
        )


type EncodedQuadraticExpression = List[Tuple[Tuple[int, int], float]]


def encode_quadratic_expression(exp: QuadraticExpression) -> EncodedQuadraticExpression:
    return cast(EncodedQuadraticExpression, list(exp.to_dict().items()))


def decode_quadratic_expression(
    exp: EncodedQuadraticExpression,
) -> Dict[Tuple[Union[int, str], Union[int, str]], float]:
    return cast(
        Dict[Tuple[Union[int, str], Union[int, str]], float],
        {key: value for key, value in exp},
    )


class EncodedQuadraticConstraint(BaseModel):
    name: str
    linear: EncodedLinearExpression
    quadratic: EncodedQuadraticExpression
    sense: str
    rhs: float

    @classmethod
    def encode(cls, constraint: QuadraticConstraint) -> EncodedQuadraticConstraint:
        return cls(
            name=constraint.name,
            linear=encode_linear_expression(constraint.linear),
            quadratic=encode_quadratic_expression(constraint.quadratic),
            sense=constraint.sense.name,
            rhs=constraint.rhs,
        )

    def decode(self, qp: QuadraticProgram) -> QuadraticConstraint:
        return qprob.QuadraticConstraint(
            qp,
            self.name,
            decode_linear_expression(self.linear),
            decode_quadratic_expression(self.quadratic),
            qprob.constraint.ConstraintSense.convert(self.sense),
            self.rhs,
        )


class EncodedQuadraticObjective(BaseModel):
    constant: float
    linear: EncodedLinearExpression
    quadratic: EncodedQuadraticExpression
    sense: ObjSense

    @classmethod
    def encode(cls, qo: QuadraticObjective) -> EncodedQuadraticObjective:
        return cls(
            constant=qo.constant,
            linear=encode_linear_expression(qo.linear),
            quadratic=encode_quadratic_expression(qo.quadratic),
            sense=qo.sense,
        )

    def decode(self, qp: QuadraticProgram) -> QuadraticObjective:
        return qprob.QuadraticObjective(
            qp,
            self.constant,
            decode_linear_expression(self.linear),
            decode_quadratic_expression(self.quadratic),
            self.sense,
        )


class EncodedQuadraticProgram(BaseModel):
    name: str
    status: QuadraticProgramStatus
    variables: List[EncodedVariable]
    variables_index: Dict[str, int]
    linear_constraints: List[EncodedLinearConstraint]
    linear_constraints_index: Dict[str, int]
    quadratic_constraints: List[EncodedQuadraticConstraint]
    quadratic_constraints_index: Dict[str, int]
    objective: EncodedQuadraticObjective

    @classmethod
    def encode(cls, qp: QuadraticProgram) -> EncodedQuadraticProgram:
        return cls(
            name=qp.name,
            status=qp.status,
            variables=[EncodedVariable.encode(v) for v in qp.variables],
            variables_index=qp.variables_index,
            linear_constraints=[
                EncodedLinearConstraint.encode(v) for v in qp.linear_constraints
            ],
            linear_constraints_index=qp.linear_constraints_index,
            quadratic_constraints=[
                EncodedQuadraticConstraint.encode(v) for v in qp.quadratic_constraints
            ],
            quadratic_constraints_index=qp.quadratic_constraints_index,
            objective=EncodedQuadraticObjective.encode(qp.objective),
        )

    def decode(self) -> QuadraticProgram:
        qp = qprob.QuadraticProgram(self.name)
        qp._status = self.status
        qp._variables = [v.decode(qp) for v in self.variables]
        qp._variables_index = self.variables_index
        qp._linear_constraints = [v.decode(qp) for v in self.linear_constraints]
        qp._linear_constraints_index = self.linear_constraints_index
        qp._quadratic_constraints = [v.decode(qp) for v in self.quadratic_constraints]
        qp._quadratic_constraints_index = self.quadratic_constraints_index
        qp._objective = self.objective.decode(qp)
        return qp



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2020, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A converter from quadratic program to a QUBO."""

from typing import List, Optional, Union, cast

import numpy as np

from ..converters.flip_problem_sense import MaximizeToMinimize
from ..converters.inequality_to_equality import InequalityToEquality
from ..converters.integer_to_binary import IntegerToBinary
from ..converters.linear_equality_to_penalty import LinearEqualityToPenalty
from ..converters.linear_inequality_to_penalty import LinearInequalityToPenalty
from ..exceptions import QiskitOptimizationError
from ..problems.quadratic_program import QuadraticProgram
from .quadratic_program_converter import QuadraticProgramConverter


class QuadraticProgramToQubo(QuadraticProgramConverter):
    """Convert a given optimization problem to a new problem that is a QUBO.

    Examples:
        >>> from qiskit_optimization.problems import QuadraticProgram
        >>> from qiskit_optimization.converters import QuadraticProgramToQubo
        >>> problem = QuadraticProgram()
        >>> # define a problem
        >>> conv = QuadraticProgramToQubo()
        >>> problem2 = conv.convert(problem)
    """

    def __init__(self, penalty: Optional[float] = None) -> None:
        """
        Args:
            penalty: Penalty factor to scale equality constraints that are added to objective.
                If None is passed, a penalty factor will be automatically calculated on every
                conversion.
        """
        self._penalize_lin_eq_constraints = LinearEqualityToPenalty(penalty=penalty)
        self._penalize_lin_ineq_constraints = LinearInequalityToPenalty(penalty=penalty)
        self._converters = [
            self._penalize_lin_ineq_constraints,
            InequalityToEquality(mode="integer"),
            IntegerToBinary(),
            self._penalize_lin_eq_constraints,
            MaximizeToMinimize(),
        ]

    def convert(self, problem: QuadraticProgram) -> QuadraticProgram:
        """Convert a problem with linear constraints into new one with a QUBO form.

        Args:
            problem: The problem with linear constraints to be solved.

        Returns:
            The problem converted in QUBO format as minimization problem.

        Raises:
            QiskitOptimizationError: In case of an incompatible problem.
        """

        # analyze compatibility of problem
        msg = self.get_compatibility_msg(problem)
        if len(msg) > 0:
            raise QiskitOptimizationError(f"Incompatible problem: {msg}")

        for conv in self._converters:
            problem = conv.convert(problem)
        return problem

    def interpret(self, x: Union[np.ndarray, List[float]]) -> np.ndarray:
        """Convert a result of a converted problem into that of the original problem.

        Args:
            x: The result of the converted problem.

        Returns:
            The result of the original problem.
        """
        for conv in self._converters[::-1]:
            x = conv.interpret(x)
        return cast(np.ndarray, x)

    @staticmethod
    def get_compatibility_msg(problem: QuadraticProgram) -> str:
        """Checks whether a given problem can be solved with this optimizer.

        Checks whether the given problem is compatible, i.e., whether the problem can be converted
        to a QUBO, and otherwise, returns a message explaining the incompatibility.

        Args:
            problem: The optimization problem to check compatibility.

        Returns:
            A message describing the incompatibility.
        """

        # initialize message
        msg = ""
        # check whether there are incompatible variable types
        if problem.get_num_continuous_vars() > 0:
            msg += "Continuous variables are not supported! "

        # check whether there are incompatible constraint types
        if len(problem.quadratic_constraints) > 0:
            msg += "Quadratic constraints are not supported. "
        # check whether there are float coefficients in constraints
        compatible_with_integer_slack = True
        for l_constraint in problem.linear_constraints:
            linear = l_constraint.linear.to_dict()
            if any(
                isinstance(coef, float) and not coef.is_integer()
                for coef in linear.values()
            ):
                compatible_with_integer_slack = False
        for q_constraint in problem.quadratic_constraints:
            linear = q_constraint.linear.to_dict()
            quadratic = q_constraint.quadratic.to_dict()
            if any(
                isinstance(coef, float) and not coef.is_integer()
                for coef in quadratic.values()
            ) or any(
                isinstance(coef, float) and not coef.is_integer()
                for coef in linear.values()
            ):
                compatible_with_integer_slack = False
        if not compatible_with_integer_slack:
            msg += (
                "Can not convert inequality constraints to equality constraint because \
                    float coefficients are in constraints. "
            )

        # if an error occurred, return error message, otherwise, return the empty string
        return msg

    def is_compatible(self, problem: QuadraticProgram) -> bool:
        """Checks whether a given problem can be solved with the optimizer implementing this method.

        Args:
            problem: The optimization problem to check compatibility.

        Returns:
            Returns True if the problem is compatible, False otherwise.
        """
        return len(self.get_compatibility_msg(problem)) == 0

    @property
    def penalty(self) -> Optional[float]:
        """Returns the penalty factor used in conversion.

        Returns:
            The penalty factor used in conversion.
        """
        return self._penalize_lin_eq_constraints.penalty

    @penalty.setter
    def penalty(self, penalty: Optional[float]) -> None:
        """Set a new penalty factor.

        Args:
            penalty: The new penalty factor.
                     If None is passed, a penalty factor will be automatically calculated on every
                     conversion.
        """
        self._penalize_lin_ineq_constraints.penalty = penalty
        self._penalize_lin_eq_constraints.penalty = penalty



from typing import Type, TypeVar, Optional
from contextlib import AbstractContextManager
import json
import time
from io import IOBase
from pydantic import BaseModel

from httpx import (
    BasicAuth,
    URL,
    Headers,
    Request,
    Client as BaseClient,
    Response,
)

from q3as.api import JobRequest, JobInfo, JobStatus, Job


class Credentials:
    """
    Credentials for authenticating with the server.
    """

    id: str
    secret: str

    def __init__(self, id: str, secret: str):
        """
        Create a new credentials instance.
        """
        self.id = id
        self.secret = secret

    @classmethod
    def load(cls, file: str | IOBase):
        """
        Load credentials from a file.
        """
        if isinstance(file, str):
            file = open(file)
        return cls(**json.load(file))

    def auth(self):
        return BasicAuth(self.id, self.secret)


class RequestBuilder:
    url: URL

    def __init__(self, url: str):
        self.url = URL(url)

    def build_request(
        self, method: str, path: str, headers: dict[str, str] = {}, **kwargs
    ) -> Request:
        return Request(
            method,
            self.url.copy_with(path=path),
            headers=Headers(
                {
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                    **headers,
                }
            ),
            **kwargs,
        )

    def create_job(self, job: JobRequest) -> Request:
        return self.build_request(
            "POST", "/api/v1/jobs/", content=job.model_dump_json()
        )

    def get_job(self, slug: str) -> Request:
        return self.build_request("GET", f"/api/v1/jobs/{slug}")

    def pause_job(self, slug: str) -> Request:
        return self.build_request("PUT", f"/api/v1/jobs/{slug}/pause")

    def resume_job(self, slug: str) -> Request:
        return self.build_request("PUT", f"/api/v1/jobs/{slug}/resume")

    def delete_job(self, slug: str) -> Request:
        return self.build_request("DELETE", f"/api/v1/jobs/{slug}")


class APIError(Exception):
    def __init__(self, message: str, response: Response):
        super().__init__(message)
        self.response = response


T = TypeVar("T")


class ResponseBuilder:
    def check(self, response: Response):
        if response.is_success:
            return
        try:
            message = response.json()["details"]
        except Exception:
            message = response.text
        raise APIError(message, response)

    def parse[T: BaseModel](self, model: Type[T], response: Response) -> T:
        self.check(response)
        return model.model_validate(response.json())


JobInput = TypeVar("JobInput")


class Client(AbstractContextManager):
    """Synchronous client for creating and managing Jobs"""

    client: BaseClient
    req: RequestBuilder
    res: ResponseBuilder

    def __init__(self, credentials: Credentials, *, url: str = "https://q3as.aqora.io"):
        """Create a new client instance"""
        self.req = RequestBuilder(url)
        self.res = ResponseBuilder()
        self.client = BaseClient(auth=credentials.auth())

    def close(self):
        """
        Close the client
        """
        self.client.close()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()

    def _send_job(self, request: Request) -> Job:
        return Job(self, self.res.parse(JobInfo, self.client.send(request)))

    def create_job(self, job: JobRequest) -> Job:
        """
        Create a new job
        """
        return self._send_job(self.req.create_job(job))

    def get_job(self, name: str) -> Job:
        """
        Get a job by its name
        """
        return self._send_job(self.req.get_job(name))

    def pause_job(self, name: str) -> Job:
        """
        Pause a job
        """
        return self._send_job(self.req.pause_job(name))

    def resume_job(self, name: str) -> Job:
        """
        Resume a job
        """
        return self._send_job(self.req.resume_job(name))

    def delete_job(self, name: str) -> Job:
        """
        Delete a job
        """
        return self._send_job(self.req.delete_job(name))

    def wait_for_job(
        self,
        name: str,
        polling_interval: float = 1.0,
        max_wait: Optional[float] = None,
    ) -> Job:
        """
        Wait for a job to finish
        """
        started = time.time()
        while True:
            job = self.get_job(name)
            if job.status is not JobStatus.STARTED:
                return job
            if max_wait is not None and time.time() - started >= max_wait:
                raise TimeoutError("Job did not finish in time")
            time.sleep(polling_interval)



from q3as.algo.vqe import VQE
from q3as.quadratic import QuadraticProgram
from q3as.app import Qubo
from q3as.encoding import EncodedVQE


def simple_qp():
    qp = QuadraticProgram("qp")

    qp.integer_var(name="x", lowerbound=-1, upperbound=1)
    qp.integer_var(name="y", lowerbound=-1, upperbound=5)
    qp.integer_var(name="z", lowerbound=-1, upperbound=5)
    qp.minimize(constant=3, linear={"x": 1}, quadratic={("x", "y"): 2, ("z", "z"): -1})

    return qp


def test_encoding_quadratic():
    vqe = VQE.builder().app(Qubo(simple_qp())).build()
    encoded = EncodedVQE.encode(vqe).model_dump_json()
    print(encoded)
    decoded = EncodedVQE.decode(EncodedVQE.model_validate_json(encoded))
    print(decoded.app.program.prettyprint())



from .numpy import EncodedArray
from .qiskit import EncodedEstimatorResult, EncodedObservables, EncodedQuantumCircuit
from .optimizer import EncodedOptimizer
from .vqe import EncodedVQE, EncodedVQEIteration, EncodedVQEResult

__all__ = [
    "EncodedArray",
    "EncodedEstimatorResult",
    "EncodedObservables",
    "EncodedQuantumCircuit",
    "EncodedOptimizer",
    "EncodedVQE",
    "EncodedVQEIteration",
    "EncodedVQEResult",
]



from __future__ import annotations
from typing import Tuple, List, Union, Dict

from pydantic import BaseModel

import numpy as np

from qiskit.quantum_info import SparsePauliOp
from qiskit.primitives.containers.observables_array import (
    ObservablesArrayLike,
)

from .application import Application, ApplicationName, BitString

type InputNode = Union[int, str]
type EdgesInput = List[
    Union[Tuple[InputNode, InputNode, float], Tuple[InputNode, InputNode]]
]


class MaxcutOutput(BaseModel):
    """
    The output of the Maxcut application
    """

    s: List[InputNode]
    "Nodes in S"
    t: List[InputNode]
    "Nodes in T"
    edges: List[Tuple[InputNode, InputNode, float]]
    "The list of edges that need to be cut"

    def __hash__(self):
        return hash(self.model_dump_json())


class Maxcut(Application[EdgesInput, MaxcutOutput]):
    def __init__(self, edges: EdgesInput):
        """
        Create a new Maxcut application. This takes in a list of edges, where each edge is a tuple of two nodes and a weight. Nodes can be either strings or integers
        """
        self.edges, self.nodes = Maxcut._normalize_edges(edges)

    @classmethod
    def _normalize_edges(
        cls, input: EdgesInput
    ) -> Tuple[Dict[Tuple[int, int], float], List[InputNode]]:
        edges = []
        nodes = set()
        for edge in input:
            if len(edge) == 3:
                new_edge = edge
            elif len(edge) == 2:
                new_edge = (edge[0], edge[1], 1.0)
            else:
                raise ValueError("Invalid edge format")
            edges.append(new_edge)
            nodes.add(new_edge[0])
            nodes.add(new_edge[1])
        nodes = list(nodes)
        remap = {x: i for i, x in enumerate(nodes)}
        return (
            {(remap[e[0]], remap[e[1]]): e[2] for e in edges},
            nodes,
        )

    def name(self) -> ApplicationName:
        return "maxcut"

    def encode(self) -> EdgesInput:
        return [(self.nodes[i], self.nodes[j], w) for (i, j), w in self.edges.items()]

    @classmethod
    def decode(cls, encoded: EdgesInput) -> Maxcut:
        return cls(encoded)

    def hamiltonian(self) -> ObservablesArrayLike:
        pauli_list = []
        for (i, j), w in self.edges.items():
            paulis = ["I"] * len(self.nodes)
            paulis[i], paulis[j] = "Z", "Z"
            pauli_list.append(("".join(paulis)[::-1], w))
        return SparsePauliOp.from_list(pauli_list)

    def interpret(self, bit_string: BitString) -> MaxcutOutput:
        """
        Interpret the bit string as a solution to the Maxcut problem. This will return the nodes in S and T, as well as the edges that need to be cut
        """
        s = set()
        t = set()
        for i, bit in enumerate(np.flip(bit_string.to_list())):
            if bit:
                s.add(i)
            else:
                t.add(i)
        edges = []
        for u in s:
            for v in t:
                w = self.edges.get((u, v), self.edges.get((v, u)))
                if w is not None:
                    edges.append((self.nodes[u], self.nodes[v], w))
        s = sorted([self.nodes[i] for i in s])
        t = sorted([self.nodes[i] for i in t])
        return MaxcutOutput(s=s, t=t, edges=edges)



from abc import ABC, abstractmethod


class Optimizer(ABC):
    """An optimizer for variational algorithms."""

    @property
    @abstractmethod
    def scipy_method(self) -> str:
        """The name of the optimizer method in scipy.optimize."""
        pass


class COBYLA(Optimizer):
    @property
    def scipy_method(self) -> str:
        return "COBYLA"


class SLSQP(Optimizer):
    @property
    def scipy_method(self) -> str:
        return "SLSQP"



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Substitute variables of QuadraticProgram."""

import logging
from collections import defaultdict
from dataclasses import dataclass
from math import isclose
from typing import Dict, Optional, Tuple, Union, cast

from ..exceptions import QiskitOptimizationError
from ..infinity import INFINITY
from .constraint import ConstraintSense
from .linear_expression import LinearExpression
from .quadratic_expression import QuadraticExpression
from .quadratic_program import QuadraticProgram

logger = logging.getLogger(__name__)


@dataclass
class SubstitutionExpression:
    """Represents a substitution of a variable with a linear expression.

    If ``variable`` is ``None``, it substitutes a variable with the constant value.
    Otherwise, it substitutes a variable with (``constant + coefficient * new_variable``).
    """

    const: float = 0.0
    """Constant value"""
    coeff: float = 0.0
    """Coefficient of the new variable"""
    variable: Optional[str] = None
    """Variable name or `None`"""


def substitute_variables(
    quadratic_program: QuadraticProgram,
    constants: Optional[Dict[Union[str, int], float]] = None,
    variables: Optional[Dict[Union[str, int], Tuple[Union[str, int], float]]] = None,
) -> QuadraticProgram:
    """Substitutes variables with constants or other variables.

    Args:
        quadratic_program: a quadratic program whose variables are substituted.

        constants: replace variable by constant
            e.g., ``{'x': 2}`` means ``x`` is substituted with 2

        variables: replace variables by weighted other variable
            need to copy everything using name reference to make sure that indices are matched
            correctly. The lower and upper bounds are updated accordingly.
            e.g., ``{'x': ('y', 2)}`` means ``x`` is substituted with ``y`` * 2

    Returns:
        An optimization problem by substituting variables with constants or other variables.
        If the substitution is valid, ``QuadraticProgram.status`` is still
        ``QuadraticProgram.Status.VALID``.
        Otherwise, it gets ``QuadraticProgram.Status.INFEASIBLE``.

    Raises:
        QiskitOptimizationError: if the substitution is invalid as follows.

            - Same variable is substituted multiple times.
            - Coefficient of variable substitution is zero.
    """
    # guarantee that there is no overlap between variables to be replaced and combine input
    subs = {}
    if constants:
        for i, v in constants.items():
            # substitute i <- v
            i_2 = quadratic_program.get_variable(i).name
            if i_2 in subs:
                raise QiskitOptimizationError(
                    f"Cannot substitute the same variable twice: {i} <- {v}"
                )
            subs[i_2] = SubstitutionExpression(const=v)

    if variables:
        for i, (j, v) in variables.items():
            if v == 0:
                raise QiskitOptimizationError(
                    f"coefficient must be non-zero: {i} {j} {v}"
                )
            # substitute i <- j * v
            i_2 = quadratic_program.get_variable(i).name
            j_2 = quadratic_program.get_variable(j).name
            if i_2 == j_2:
                raise QiskitOptimizationError(
                    f"Cannot substitute the same variable: {i} <- {j} {v}"
                )
            if i_2 in subs:
                raise QiskitOptimizationError(
                    f"Cannot substitute the same variable twice: {i} <- {j} {v}"
                )
            if j_2 in subs:
                raise QiskitOptimizationError(
                    "Cannot substitute by variable that gets substituted itself: "
                    f"{i} <- {j} {v}"
                )
            subs[i_2] = SubstitutionExpression(variable=j_2, coeff=v)

    return _SubstituteVariables().substitute_variables(quadratic_program, subs)


class _SubstituteVariables:
    """A class to substitute variables of an optimization problem with constants for other
    variables"""

    def __init__(self) -> None:
        self._src: Optional[QuadraticProgram] = None
        self._dst: Optional[QuadraticProgram] = None
        self._subs: Dict[str, SubstitutionExpression] = {}

    def substitute_variables(
        self,
        quadratic_program: QuadraticProgram,
        subs: Dict[str, SubstitutionExpression],
    ) -> QuadraticProgram:
        """Substitutes variables with constants or other variables.

        Args:
            quadratic_program: a quadratic program whose variables are substituted.

            subs: substitution expressions as a dictionary.
                e.g., {'x': SubstitutionExpression(const=1, coeff=2, variable='y'} means
                `x` is substituted with `1 + 2 * y`.

        Returns:
            An optimization problem by substituting variables with constants or other variables.
            If the substitution is valid, `QuadraticProgram.status` is still
            `QuadraticProgram.Status.VALID`.
            Otherwise, it gets `QuadraticProgram.Status.INFEASIBLE`.
        """
        self._src = quadratic_program
        self._dst = QuadraticProgram(quadratic_program.name)
        self._subs = subs
        results = [
            self._variables(),
            self._objective(),
            self._linear_constraints(),
            self._quadratic_constraints(),
        ]
        if not all(results):
            self._dst._status = QuadraticProgram.Status.INFEASIBLE
        return self._dst

    @staticmethod
    def _feasible(sense: ConstraintSense, rhs: float) -> bool:
        """Checks feasibility of the following condition
        0 `sense` rhs
        """
        if sense == ConstraintSense.EQ:
            if rhs == 0:
                return True
        elif sense == ConstraintSense.LE:
            if rhs >= 0:
                return True
        elif sense == ConstraintSense.GE:
            if rhs <= 0:
                return True
        return False

    def _variables(self) -> bool:
        # copy variables that are not replaced
        feasible = True
        for var in self._src.variables:
            name = var.name
            vartype = var.vartype
            lowerbound = var.lowerbound
            upperbound = var.upperbound
            if name not in self._subs:
                self._dst._add_variable(lowerbound, upperbound, vartype, name)

        for i, expr in self._subs.items():
            lb_i = self._src.get_variable(i).lowerbound
            ub_i = self._src.get_variable(i).upperbound
            # substitute x_i <- x_j * coeff + const
            # lb_i <= x_i <= ub_i  -->
            #   (lb_i - const) / coeff <=  x_j  <= (ub_i - const) / coeff    if coeff > 0
            #   (ub_i - const) / coeff <=  x_j  <= (lb_i - const) / coeff    if coeff < 0
            #                     lb_i <= const <= ub_i                      if coeff == 0
            if isclose(expr.coeff, 0.0, abs_tol=1e-10):
                if not lb_i <= expr.const <= ub_i:
                    logger.warning("Infeasible substitution for variable: %s", i)
                    feasible = False
            else:
                if abs(lb_i) < INFINITY:
                    new_lb_i = (lb_i - expr.const) / expr.coeff
                else:
                    new_lb_i = lb_i if expr.coeff > 0 else -lb_i
                if abs(ub_i) < INFINITY:
                    new_ub_i = (ub_i - expr.const) / expr.coeff
                else:
                    new_ub_i = ub_i if expr.coeff > 0 else -ub_i
                var_j = self._dst.get_variable(expr.variable)
                lb_j = var_j.lowerbound
                ub_j = var_j.upperbound
                if expr.coeff > 0:
                    var_j.lowerbound = max(lb_j, new_lb_i)
                    var_j.upperbound = min(ub_j, new_ub_i)
                else:
                    var_j.lowerbound = max(lb_j, new_ub_i)
                    var_j.upperbound = min(ub_j, new_lb_i)

        for var in self._dst.variables:
            if var.lowerbound > var.upperbound:
                logger.warning(
                    "Infeasible lower and upper bounds: %s %f %f",
                    var,
                    var.lowerbound,
                    var.upperbound,
                )
                feasible = False

        return feasible

    def _linear_expression(
        self, lin_expr: LinearExpression
    ) -> Tuple[float, LinearExpression]:
        const = 0.0
        lin_dict: Dict[str, float] = defaultdict(float)
        for i, w_i in lin_expr.to_dict(use_name=True).items():
            i = cast(str, i)
            expr_i = self._subs.get(i, SubstitutionExpression(coeff=1, variable=i))
            const += w_i * expr_i.const
            if expr_i.variable:
                lin_dict[expr_i.variable] += w_i * expr_i.coeff
        new_lin = LinearExpression(
            quadratic_program=self._dst, coefficients=lin_dict if lin_dict else {}
        )
        return const, new_lin

    def _quadratic_expression(
        self, quad_expr: QuadraticExpression
    ) -> Tuple[float, Optional[LinearExpression], Optional[QuadraticExpression]]:
        const = 0.0
        lin_dict: Dict[str, float] = defaultdict(float)
        quad_dict: Dict[Tuple[str, str], float] = defaultdict(float)
        for (i, j), w_ij in quad_expr.to_dict(use_name=True).items():
            i = cast(str, i)
            j = cast(str, j)
            expr_i = self._subs.get(i, SubstitutionExpression(coeff=1, variable=i))
            expr_j = self._subs.get(j, SubstitutionExpression(coeff=1, variable=j))
            const += w_ij * expr_i.const * expr_j.const
            if expr_i.variable:
                lin_dict[expr_i.variable] += w_ij * expr_i.coeff * expr_j.const
            if expr_j.variable:
                lin_dict[expr_j.variable] += w_ij * expr_j.coeff * expr_i.const
            if expr_i.variable and expr_j.variable:
                quad_dict[expr_i.variable, expr_j.variable] += (
                    w_ij * expr_i.coeff * expr_j.coeff
                )
        new_lin = LinearExpression(
            quadratic_program=self._dst, coefficients=lin_dict if lin_dict else {}
        )
        new_quad = QuadraticExpression(
            quadratic_program=self._dst, coefficients=quad_dict if quad_dict else {}
        )
        return const, new_lin, new_quad

    def _objective(self) -> bool:
        obj = self._src.objective
        const1, lin1 = self._linear_expression(obj.linear)
        const2, lin2, quadratic = self._quadratic_expression(obj.quadratic)

        constant = obj.constant + const1 + const2
        linear = lin1.coefficients + lin2.coefficients
        if obj.sense == obj.sense.MINIMIZE:
            self._dst.minimize(
                constant=constant, linear=linear, quadratic=quadratic.coefficients
            )
        else:
            self._dst.maximize(
                constant=constant, linear=linear, quadratic=quadratic.coefficients
            )
        return True

    def _linear_constraints(self) -> bool:
        feasible = True
        for lin_cst in self._src.linear_constraints:
            constant, linear = self._linear_expression(lin_cst.linear)
            rhs = lin_cst.rhs - constant
            if linear.coefficients.nnz > 0:
                self._dst.linear_constraint(
                    name=lin_cst.name,
                    linear=linear.coefficients,
                    sense=lin_cst.sense,
                    rhs=rhs,
                )
            else:
                if not self._feasible(lin_cst.sense, rhs):
                    logger.warning(
                        "constraint %s is infeasible due to substitution", lin_cst.name
                    )
                    feasible = False
        return feasible

    def _quadratic_constraints(self) -> bool:
        feasible = True
        for quad_cst in self._src.quadratic_constraints:
            const1, lin1 = self._linear_expression(quad_cst.linear)
            const2, lin2, quadratic = self._quadratic_expression(quad_cst.quadratic)
            rhs = quad_cst.rhs - const1 - const2
            linear = lin1.coefficients + lin2.coefficients

            if quadratic.coefficients.nnz > 0:
                self._dst.quadratic_constraint(
                    name=quad_cst.name,
                    linear=linear,
                    quadratic=quadratic.coefficients,
                    sense=quad_cst.sense,
                    rhs=rhs,
                )
            elif linear.nnz > 0:
                name = quad_cst.name
                lin_names = set(lin.name for lin in self._dst.linear_constraints)
                while name in lin_names:
                    name = "_" + name
                self._dst.linear_constraint(
                    name=name, linear=linear, sense=quad_cst.sense, rhs=rhs
                )
            else:
                if not self._feasible(quad_cst.sense, rhs):
                    logger.warning(
                        "constraint %s is infeasible due to substitution", quad_cst.name
                    )
                    feasible = False
        return feasible



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2020, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""The converter to map integer variables in a quadratic program to binary variables."""

import copy
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from .quadratic_program_converter import QuadraticProgramConverter
from ..exceptions import QiskitOptimizationError
from ..problems.quadratic_objective import QuadraticObjective
from ..problems.quadratic_program import QuadraticProgram
from ..problems.variable import Variable


class IntegerToBinary(QuadraticProgramConverter):
    """Convert a :class:`~qiskit_optimization.problems.QuadraticProgram` into new one by encoding
    integer with binary variables.

    This bounded-coefficient encoding used in this converted is proposed in [1], Eq. (5).

    Examples:
        >>> from qiskit_optimization.problems import QuadraticProgram
        >>> from qiskit_optimization.converters import IntegerToBinary
        >>> problem = QuadraticProgram()
        >>> var = problem.integer_var(name='x', lowerbound=0, upperbound=10)
        >>> conv = IntegerToBinary()
        >>> problem2 = conv.convert(problem)

    References:
        [1]: Sahar Karimi, Pooya Ronagh (2017), Practical Integer-to-Binary Mapping for Quantum
            Annealers. arxiv.org:1706.01945.
    """

    _delimiter = "@"  # users are supposed not to use this character in variable names

    def __init__(self) -> None:
        self._src: Optional[QuadraticProgram] = None
        self._dst: Optional[QuadraticProgram] = None
        self._conv: Dict[Variable, List[Tuple[str, int]]] = {}
        # e.g., self._conv = {x: [('x@1', 1), ('x@2', 2)]}

    def convert(self, problem: QuadraticProgram) -> QuadraticProgram:
        """Convert an integer problem into a new problem with binary variables.

        Args:
            problem: The problem to be solved, that may contain integer variables.

        Returns:
            The converted problem, that contains no integer variables.

        Raises:
            QiskitOptimizationError: if variable or constraint type is not supported.
        """

        # Copy original QP as reference.
        self._src = copy.deepcopy(problem)

        if self._src.get_num_integer_vars() > 0:
            # Initialize new QP
            self._dst = QuadraticProgram(name=problem.name)

            # Declare variables
            for x in self._src.variables:
                if x.vartype == Variable.Type.INTEGER:
                    new_vars = self._convert_var(x.name, x.lowerbound, x.upperbound)
                    self._conv[x] = new_vars
                    for var_name, _ in new_vars:
                        self._dst.binary_var(var_name)
                else:
                    if x.vartype == Variable.Type.CONTINUOUS:
                        self._dst.continuous_var(x.lowerbound, x.upperbound, x.name)
                    elif x.vartype == Variable.Type.BINARY:
                        self._dst.binary_var(x.name)
                    else:
                        raise QiskitOptimizationError(
                            f"Unsupported variable type {x.vartype}"
                        )

            self._substitute_int_var()

        else:
            # just copy the problem if no integer variables exist
            self._dst = copy.deepcopy(problem)

        return self._dst

    def _convert_var(
        self, name: str, lowerbound: float, upperbound: float
    ) -> List[Tuple[str, int]]:
        var_range = upperbound - lowerbound
        power = int(np.log2(var_range)) if var_range > 0 else 0
        bounded_coef = var_range - (2**power - 1)

        coeffs = [2**i for i in range(power)] + [bounded_coef]
        return [
            (name + self._delimiter + str(i), coef) for i, coef in enumerate(coeffs)
        ]

    def _convert_linear_coefficients_dict(
        self, coefficients: Dict[str, float]
    ) -> Tuple[Dict[str, float], float]:
        constant = 0.0
        linear: Dict[str, float] = {}
        for name, v in coefficients.items():
            x = self._src.get_variable(name)
            if x in self._conv:
                for y, coeff in self._conv[x]:
                    linear[y] = v * coeff
                constant += v * x.lowerbound
            else:
                linear[x.name] = v

        return linear, constant

    def _convert_quadratic_coefficients_dict(
        self, coefficients: Dict[Tuple[str, str], float]
    ) -> Tuple[Dict[Tuple[str, str], float], Dict[str, float], float]:
        constant = 0.0
        linear: Dict[str, float] = {}
        quadratic = {}
        for (name_i, name_j), v in coefficients.items():
            x = self._src.get_variable(name_i)
            y = self._src.get_variable(name_j)

            if x in self._conv and y not in self._conv:
                for z_x, coeff_x in self._conv[x]:
                    quadratic[z_x, y.name] = v * coeff_x
                linear[y.name] = linear.get(y.name, 0.0) + v * x.lowerbound

            elif x not in self._conv and y in self._conv:
                for z_y, coeff_y in self._conv[y]:
                    quadratic[x.name, z_y] = v * coeff_y
                linear[x.name] = linear.get(x.name, 0.0) + v * y.lowerbound

            elif x in self._conv and y in self._conv:
                for z_x, coeff_x in self._conv[x]:
                    for z_y, coeff_y in self._conv[y]:
                        quadratic[z_x, z_y] = v * coeff_x * coeff_y

                for z_x, coeff_x in self._conv[x]:
                    linear[z_x] = linear.get(z_x, 0.0) + v * coeff_x * y.lowerbound
                for z_y, coeff_y in self._conv[y]:
                    linear[z_y] = linear.get(z_y, 0.0) + v * coeff_y * x.lowerbound

                constant += v * x.lowerbound * y.lowerbound

            else:
                quadratic[x.name, y.name] = v

        return quadratic, linear, constant

    def _substitute_int_var(self):
        # set objective
        linear, linear_constant = self._convert_linear_coefficients_dict(
            self._src.objective.linear.to_dict(use_name=True)
        )
        (
            quadratic,
            q_linear,
            q_constant,
        ) = self._convert_quadratic_coefficients_dict(
            self._src.objective.quadratic.to_dict(use_name=True)
        )

        constant = self._src.objective.constant + linear_constant + q_constant
        for i, v in q_linear.items():
            linear[i] = linear.get(i, 0) + v

        if self._src.objective.sense == QuadraticObjective.Sense.MINIMIZE:
            self._dst.minimize(constant, linear, quadratic)
        else:
            self._dst.maximize(constant, linear, quadratic)

        # set linear constraints
        for constraint in self._src.linear_constraints:
            linear, constant = self._convert_linear_coefficients_dict(
                constraint.linear.to_dict(use_name=True)
            )
            self._dst.linear_constraint(
                linear, constraint.sense, constraint.rhs - constant, constraint.name
            )

        # set quadratic constraints
        for constraint in self._src.quadratic_constraints:
            linear, linear_constant = self._convert_linear_coefficients_dict(
                constraint.linear.to_dict(use_name=True)
            )
            quadratic, q_linear, q_constant = self._convert_quadratic_coefficients_dict(
                constraint.quadratic.to_dict(use_name=True)
            )

            constant = linear_constant + q_constant
            for i, v in q_linear.items():
                linear[i] = linear.get(i, 0) + v

            self._dst.quadratic_constraint(
                linear,
                quadratic,
                constraint.sense,
                constraint.rhs - constant,
                constraint.name,
            )

    def interpret(self, x: Union[np.ndarray, List[float]]) -> np.ndarray:
        """Convert back the converted problem (binary variables)
        to the original (integer variables).

        Args:
            x: The result of the converted problem or the given result in case of FAILURE.

        Returns:
            The result of the original problem.
        """
        # interpret integer values
        sol = {var.name: x[i] for i, var in enumerate(self._dst.variables)}
        new_x = np.zeros(self._src.get_num_vars())
        for i, var in enumerate(self._src.variables):
            if var in self._conv:
                new_x[i] = (
                    sum(sol[aux] * coef for aux, coef in self._conv[var])
                    + var.lowerbound
                )
            else:
                new_x[i] = sol[var.name]
        return np.array(new_x)



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Quadratic Objective."""

from typing import Union, List, Dict, Tuple, Any, Optional

from numpy import ndarray
from scipy.sparse import spmatrix

from ..exceptions import QiskitOptimizationError
from .linear_constraint import LinearExpression
from .quadratic_expression import QuadraticExpression
from .quadratic_program_element import QuadraticProgramElement
from .types import ObjSense


class QuadraticObjective(QuadraticProgramElement):
    """Representation of quadratic objective function of the form:
    constant + linear * x + x * quadratic * x.
    """

    Sense = ObjSense

    def __init__(
        self,
        quadratic_program: Any,
        constant: float = 0.0,
        linear: Optional[
            Union[ndarray, spmatrix, List[float], Dict[Union[str, int], float]]
        ] = None,
        quadratic: Optional[
            Union[
                ndarray,
                spmatrix,
                List[List[float]],
                Dict[Tuple[Union[int, str], Union[int, str]], float],
            ]
        ] = None,
        sense: ObjSense = ObjSense.MINIMIZE,
    ) -> None:
        """Constructs a quadratic objective function.

        Args:
            quadratic_program: The parent quadratic program.
            constant: The constant offset of the objective.
            linear: The coefficients of the linear part of the objective.
            quadratic: The coefficients of the quadratic part of the objective.
            sense: The optimization sense of the objective.
        """
        super().__init__(quadratic_program)
        self._constant = constant
        if linear is None:
            linear = {}
        self._linear = LinearExpression(quadratic_program, linear)
        if quadratic is None:
            quadratic = {}
        self._quadratic = QuadraticExpression(quadratic_program, quadratic)
        self._sense = sense

    @property
    def constant(self) -> float:
        """Returns the constant part of the objective function.

        Returns:
            The constant part of the objective function.
        """
        return self._constant

    @constant.setter
    def constant(self, constant: float) -> None:
        """Sets the constant part of the objective function.

        Args:
            constant: The constant part of the objective function.
        """
        self._constant = constant

    @property
    def linear(self) -> LinearExpression:
        """Returns the linear part of the objective function.

        Returns:
            The linear part of the objective function.
        """
        return self._linear

    @linear.setter
    def linear(
        self,
        linear: Union[ndarray, spmatrix, List[float], Dict[Union[str, int], float]],
    ) -> None:
        """Sets the coefficients of the linear part of the objective function.

        Args:
            linear: The coefficients of the linear part of the objective function.

        """
        self._linear = LinearExpression(self.quadratic_program, linear)

    @property
    def quadratic(self) -> QuadraticExpression:
        """Returns the quadratic part of the objective function.

        Returns:
            The quadratic part of the objective function.
        """
        return self._quadratic

    @quadratic.setter
    def quadratic(
        self,
        quadratic: Union[
            ndarray,
            spmatrix,
            List[List[float]],
            Dict[Tuple[Union[int, str], Union[int, str]], float],
        ],
    ) -> None:
        """Sets the coefficients of the quadratic part of the objective function.

        Args:
            quadratic: The coefficients of the quadratic part of the objective function.

        """
        self._quadratic = QuadraticExpression(self.quadratic_program, quadratic)

    @property
    def sense(self) -> ObjSense:
        """Returns the sense of the objective function.

        Returns:
            The sense of the objective function.
        """
        return self._sense

    @sense.setter
    def sense(self, sense: ObjSense) -> None:
        """Sets the sense of the objective function.

        Args:
            sense: The sense of the objective function.
        """
        self._sense = sense

    def evaluate(self, x: Union[ndarray, List, Dict[Union[int, str], float]]) -> float:
        """Evaluate the quadratic objective for given variable values.

        Args:
            x: The values of the variables to be evaluated.

        Returns:
            The value of the quadratic objective given the variable values.

        Raises:
            QiskitOptimizationError: if the shape of the objective function does not match with
                the number of variables.
        """
        n = self.quadratic_program.get_num_vars()
        if self.linear.coefficients.shape != (
            1,
            n,
        ) or self.quadratic.coefficients.shape != (n, n):
            raise QiskitOptimizationError(
                "The shape of the objective function does not match with the number of variables. "
                "Need to define the objective function after defining all variables"
            )
        return self.constant + self.linear.evaluate(x) + self.quadratic.evaluate(x)

    def evaluate_gradient(
        self, x: Union[ndarray, List, Dict[Union[int, str], float]]
    ) -> ndarray:
        """Evaluate the gradient of the quadratic objective for given variable values.

        Args:
            x: The values of the variables to be evaluated.

        Returns:
            The value of the gradient of the quadratic objective given the variable values.

        Raises:
            QiskitOptimizationError: if the shape of the objective function does not match with
                the number of variables.
        """
        n = self.quadratic_program.get_num_vars()
        if self.linear.coefficients.shape != (
            1,
            n,
        ) or self.quadratic.coefficients.shape != (n, n):
            raise QiskitOptimizationError(
                "The shape of the objective function does not match with the number of variables. "
                "Need to define the objective function after defining all variables"
            )
        return self.linear.evaluate_gradient(x) + self.quadratic.evaluate_gradient(x)

    def __repr__(self):
        # pylint: disable=cyclic-import
        from ..translators.prettyprint import expr2str, DEFAULT_TRUNCATE

        expr_str = expr2str(
            self.constant, self.linear, self.quadratic, DEFAULT_TRUNCATE
        )
        return f"<{self.__class__.__name__}: {self._sense.name.lower()} {expr_str}>"

    def __str__(self):
        # pylint: disable=cyclic-import
        from ..translators.prettyprint import expr2str

        expr_str = expr2str(self.constant, self.linear, self.quadratic)
        return f"{self._sense.name.lower()} {expr_str}"



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Optimization problems (:mod:`qiskit_optimization.problems`)
===========================================================

.. currentmodule:: qiskit_optimization.problems

Quadratic program
-----------------
Structures for defining an optimization problem.

Note:
    The following classes are not intended to be instantiated directly.
    Objects of these types are available within an instantiated
    :class:`~qiskit_optimization.QuadraticProgram`.

.. autosummary::
   :toctree: ../stubs/
   :nosignatures:

   Constraint
   LinearExpression
   LinearConstraint
   QuadraticExpression
   QuadraticConstraint
   QuadraticObjective
   QuadraticProgramElement
   Variable

"""

from .constraint import Constraint
from .linear_constraint import LinearConstraint
from .linear_expression import LinearExpression
from .quadratic_constraint import QuadraticConstraint
from .quadratic_expression import QuadraticExpression
from .quadratic_objective import QuadraticObjective
from .quadratic_program import QuadraticProgram
from .quadratic_program_element import QuadraticProgramElement
from .variable import Variable
from .types import VarType, QuadraticProgramStatus

__all__ = [
    "Constraint",
    "LinearExpression",
    "LinearConstraint",
    "QuadraticExpression",
    "QuadraticConstraint",
    "QuadraticObjective",
    "QuadraticProgram",
    "QuadraticProgramElement",
    "Variable",
    "VarType",
    "QuadraticProgramStatus",
]



from enum import Enum


class QuadraticProgramStatus(Enum):
    """Status of QuadraticProgram"""

    VALID = 0
    INFEASIBLE = 1


class VarType(Enum):
    """Constants defining variable type."""

    CONTINUOUS = 0
    BINARY = 1
    INTEGER = 2


class ObjSense(Enum):
    """Objective Sense Type."""

    MINIMIZE = 1
    MAXIMIZE = -1



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    cast,
    Optional,
    Union,
    Callable,
    Any,
    List,
    Tuple,
    Dict,
)
from dataclasses import dataclass

import numpy as np
from numpy.typing import ArrayLike

from scipy.optimize import minimize

from qiskit.circuit import QuantumCircuit
from qiskit.primitives import (
    BaseEstimatorV2 as Estimator,
    BaseSamplerV2 as Sampler,
    PrimitiveResult,
)
from qiskit.primitives.containers.observables_array import (
    ObservablesArray,
    ObservablesArrayLike,
)
from qiskit.primitives.containers.bindings_array import BindingsArray, BindingsArrayLike
from qiskit.primitives.containers.estimator_pub import EstimatorPub
from qiskit.primitives.containers.sampler_pub_result import SamplerPubResult
from qiskit.primitives.containers.pub_result import PubResult
from qiskit.quantum_info import SparsePauliOp
from qiskit.circuit.library import (
    QAOAAnsatz,
    TwoLocal as TwoLocalAnsatz,
    EfficientSU2 as EfficientSU2Ansatz,
)

from q3as.app import Application
from q3as.run_options import RunOptions
import q3as.encoding
import q3as.api

from .optimizer import Optimizer, COBYLA
from .types import HaltReason, EstimatorData, SamplerData

if TYPE_CHECKING:
    from q3as.client import Client
    from q3as.api import Job


@dataclass
class VQEIteration:
    """
    The result of a an iteration of VQE
    """

    iter: int
    "The number of the iteration"
    cost: float
    "The value of the cost function of the iteration"
    params: np.ndarray
    "The parameters of the ansatz of the iteration"
    estimated: PrimitiveResult[PubResult]
    "The result of the estimation"
    best: bool = False
    "Whether this iteration is the best so far"


@dataclass
class VQEResult:
    params: np.ndarray
    "The parameters of the best iteration"
    iter: int = 0
    "The number of iterations"
    reason: HaltReason = HaltReason.INTERRUPT
    "The reason why the optimization stopped"
    cost: Optional[float] = None
    "The value of the cost function of the best iteration"
    estimated: Optional[PrimitiveResult[PubResult]] = None
    "The result of the estimation of the best iteration"
    sampled: Optional[PrimitiveResult[SamplerPubResult]] = None
    "The result of the sampling"
    meas_counts: Optional[Dict[str, int]] = None
    "The bit string counts after sampling"
    interpreted: Optional[List[Tuple[Any, int]]] = None
    "The interpreted results of the sampling as defined by the application if given"

    def most_sampled(self) -> Optional[Any]:
        """
        Get the the most sampled value from the results.
        This is the interpreted value as defined by the application if given or otherwise the bitstring
        """
        samples = None
        if self.interpreted is not None:
            samples = [(count, value) for value, count in self.interpreted]
        elif self.meas_counts is not None:
            samples = [
                (count, bitstring) for bitstring, count in self.meas_counts.items()
            ]
        max_sample = max(samples, key=lambda x: x[0]) if samples is not None else None
        if max_sample is not None:
            return max_sample[1]
        return None


@dataclass
class VQE:
    """
    Parameters for running VQE
    """

    ansatz: QuantumCircuit
    "The ansatz circuit"
    observables: ObservablesArray
    "The observables to estimate. E.g. the Hamiltonian of the system"
    initial_params: np.ndarray
    "The initial parameters of the ansatz"
    optimizer: Optimizer
    "The optimizer to use"
    app: Optional[Application]
    "The application to use for interpretation of results"
    maxiter: int = 1000
    "The maximum number of iterations"

    @classmethod
    def builder(cls) -> VQEBuilder:
        """
        Create a builder for VQE
        """
        return VQEBuilder()

    def run(
        self,
        estimator: Estimator,
        sampler: Optional[Sampler] = None,
        callback: Optional[Callable[[VQEIteration], None]] = None,
    ) -> VQEResult:
        """
        Run VQE
        """

        out = VQEResult(params=self.initial_params)

        def cost_fun(
            params: np.ndarray,
        ):
            out.iter += 1
            bound_params = BindingsArray.coerce({tuple(self.ansatz.parameters): params})
            pub = EstimatorPub(self.ansatz, self.observables, bound_params)
            result = estimator.run([pub]).result()
            cost = cast(EstimatorData, result[0].data).evs
            vqe_result = VQEIteration(
                iter=out.iter, cost=cost, params=params, estimated=result
            )
            if out.cost is None or cost < out.cost:
                vqe_result.best = True
                out.params = params
                out.cost = cost
                out.estimated = vqe_result.estimated
            if callback is not None:
                callback(vqe_result)
            return cost

        try:
            res = minimize(
                cost_fun,
                self.initial_params,
                method=self.optimizer.scipy_method,
                options={"maxiter": self.maxiter},
            )
            if res.success:
                out.reason = HaltReason.TOLERANCE
            else:
                out.reason = HaltReason.MAXITER
        except StopIteration:
            out.reason = HaltReason.INTERRUPT

        if sampler is not None:
            bound_params = cast(
                BindingsArrayLike,
                BindingsArray.coerce({tuple(self.ansatz.parameters): out.params}).data,
            )
            measured_ansatz = cast(
                QuantumCircuit, self.ansatz.measure_all(inplace=False)
            )
            sampled = sampler.run([(measured_ansatz, bound_params)]).result()

            out.sampled = sampled

            meas = cast(SamplerData, sampled[0].data).meas
            out.meas_counts = meas.get_counts()

            if self.app is not None:
                out.interpreted = self.app.interpreted_meas(meas)

        return out

    def send(self, api: Client, run_options: RunOptions = RunOptions()) -> Job:
        """
        Send the VQE job to the API
        """
        return api.create_job(
            q3as.api.JobRequest(
                input=q3as.encoding.EncodedVQE.encode(self), run_options=run_options
            )
        )


def _observables_to_pauli(observables: ObservablesArray) -> SparsePauliOp:
    return SparsePauliOp.from_list(observables.ravel().tolist()[0].items())


type _PostProcessCallable = Optional[Callable[[QuantumCircuit], QuantumCircuit]]


type AnsatzCallback = Callable[[ObservablesArray], QuantumCircuit]


class QAOA:
    """
    A builder for the QAOA ansatz
    """

    def __init__(self, postprocess: _PostProcessCallable = None, **kwargs):
        """
        Create a new QAOA builder. Arguments are passed to the QAOAAnsatz constructor in the Qiskit circuit library.
        """
        self.postprocess = postprocess
        self.kwargs = kwargs

    def __call__(self, obs: ObservablesArray) -> QuantumCircuit:
        ansatz = QAOAAnsatz(_observables_to_pauli(obs), **self.kwargs)
        if self.postprocess is not None:
            ansatz = self.postprocess(ansatz)
        return ansatz


class TwoLocal:
    """
    A builder for the TwoLocal ansatz
    """

    def __init__(self, postprocess: _PostProcessCallable = None, **kwargs):
        """
        Create a new TwoLocal builder. Arguments are passed to the TwoLocal constructor in the Qiskit circuit library.
        """
        self.postprocess = postprocess
        self.kwargs = kwargs

    def __call__(self, obs: ObservablesArray) -> QuantumCircuit:
        ansatz = TwoLocalAnsatz(_observables_to_pauli(obs).num_qubits, **self.kwargs)
        if self.postprocess is not None:
            ansatz = self.postprocess(ansatz)
        return ansatz


class EfficientSU2:
    """
    A builder for the EfficientSU2 ansatz
    """

    def __init__(self, postprocess: _PostProcessCallable = None, **kwargs):
        """
        Create a new EfficientSU2 builder. Arguments are passed to the EfficientSU2 constructor in the Qiskit circuit library.
        """
        self.postprocess = postprocess
        self.kwargs = kwargs

    def __call__(self, obs: ObservablesArray) -> QuantumCircuit:
        ansatz = EfficientSU2Ansatz(
            _observables_to_pauli(obs).num_qubits, **self.kwargs
        )
        if self.postprocess is not None:
            ansatz = self.postprocess(ansatz)
        return ansatz


class VQEBuilder:
    _app: Optional[Application] = None
    _ansatz: Optional[Union[QuantumCircuit, AnsatzCallback]] = EfficientSU2()
    _observables: Optional[ObservablesArrayLike] = None
    _initial_params: Optional[ArrayLike] = None
    _optimizer: Optimizer = COBYLA()
    _maxiter: int = 1000

    def app(self, app: Application):
        """
        Set the `Application` to use for interpretation of results
        """
        self._app = app
        return self

    def ansatz(
        self,
        qc: Union[QuantumCircuit, AnsatzCallback],
    ) -> VQEBuilder:
        """
        Set the ansatz to use for the VQE.
        This can either be a QuantumCircuit or a callback that takes the observables as argument and returns a QuantumCircuit.
        This defaults to the EfficientSU2 ansatz.
        """
        self._ansatz = qc
        return self

    def observables(self, obs: ObservablesArrayLike) -> VQEBuilder:
        """
        Set the observables to estimate. E.g. the Hamiltonian of the system.
        If no observables are given, the Hamiltonian of the application is used
        """
        self._observables = obs
        return self

    def initial_params(self, params: ArrayLike) -> VQEBuilder:
        """
        Set the initial parameters of the ansatz. If no parameters are given, the initial parameters are set to 0.
        """
        self._initial_params = params
        return self

    def optimizer(self, opt: Optimizer) -> VQEBuilder:
        """
        Set the optimizer to use for the VQE. This defaults to COBYLA.
        """
        self._optimizer = opt
        return self

    def maxiter(self, maxiter: int) -> VQEBuilder:
        """
        Set the maximum number of iterations for the VQE. This defaults to 1000.
        """
        self._maxiter = maxiter
        return self

    def build(self) -> VQE:
        """
        Build the VQE instance
        """
        if self._observables is None:
            if self._app is not None:
                observables = self._app.hamiltonian()
            else:
                raise ValueError("Observables are required")
        else:
            observables = self._observables
        observables = ObservablesArray.coerce(observables)

        if self._ansatz is None:
            raise ValueError("Ansatz is required")
        if isinstance(self._ansatz, QuantumCircuit):
            ansatz = self._ansatz
        else:
            ansatz = self._ansatz(observables)

        if self._initial_params is not None:
            _initial_params = np.array(self._initial_params)
            if len(_initial_params) != ansatz.num_parameters:
                raise ValueError("Initial parameters must match ansatz")
        else:
            _initial_params = np.zeros(ansatz.num_parameters)

        return VQE(
            app=self._app,
            ansatz=ansatz,
            observables=observables,
            initial_params=_initial_params,
            optimizer=self._optimizer,
            maxiter=self._maxiter,
        )

    def run(
        self,
        estimator: Estimator,
        sampler: Optional[Sampler] = None,
        callback: Optional[Callable[[VQEIteration], None]] = None,
    ) -> VQEResult:
        """
        Run the VQE
        """
        return self.build().run(estimator, sampler, callback)

    def send(self, api: Client, estimator: RunOptions = RunOptions()) -> Job:
        """
        Send the VQE job to the API
        """
        return self.build().send(api, estimator)



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Variable interface"""

from typing import Tuple, Union, Any

from .quadratic_program_element import QuadraticProgramElement
from ..exceptions import QiskitOptimizationError
from ..infinity import INFINITY
from .types import VarType


class Variable(QuadraticProgramElement):
    """Representation of a variable."""

    Type = VarType

    def __init__(
        self,
        quadratic_program: Any,
        name: str,
        lowerbound: Union[float, int] = 0,
        upperbound: Union[float, int] = INFINITY,
        vartype: VarType = VarType.CONTINUOUS,
    ) -> None:
        """Creates a new Variable.

        The variables is exposed by the top-level `QuadraticProgram` class
        in `QuadraticProgram.variables`.  This constructor is not meant to be used
        externally.

        Args:
            quadratic_program: The parent QuadraticProgram.
            name: The variable name.
            lowerbound: The variable lowerbound.
            upperbound: The variable upperbound.
            vartype: The variable type.

        Raises:
            QiskitOptimizationError: if lowerbound is greater than upperbound.
        """
        if lowerbound > upperbound:
            raise QiskitOptimizationError("Lowerbound is greater than upperbound!")

        super().__init__(quadratic_program)
        self._name = name
        self._lowerbound = lowerbound
        self._upperbound = upperbound
        self._vartype = vartype

    @property
    def name(self) -> str:
        """Returns the name of the variable.

        Returns:
            The name of the variable.
        """
        return self._name

    @property
    def lowerbound(self) -> Union[float, int]:
        """Returns the lowerbound of the variable.

        Returns:
            The lower bound of the variable.
        """
        return self._lowerbound

    @lowerbound.setter
    def lowerbound(self, lowerbound: Union[float, int]) -> None:
        """Sets the lowerbound of the variable.

        Args:
            lowerbound: The lower bound of the variable.

        Raises:
            QiskitOptimizationError: if lowerbound is greater than upperbound.
        """
        if lowerbound > self.upperbound:
            raise QiskitOptimizationError("Lowerbound is greater than upperbound!")
        self._lowerbound = lowerbound

    @property
    def upperbound(self) -> Union[float, int]:
        """Returns the upperbound of the variable.

        Returns:
            The upperbound of the variable.
        """
        return self._upperbound

    @upperbound.setter
    def upperbound(self, upperbound: Union[float, int]) -> None:
        """Sets the upperbound of the variable.

        Args:
            upperbound: The upperbound of the variable.

        Raises:
            QiskitOptimizationError: if upperbound is smaller than lowerbound.
        """
        if self.lowerbound > upperbound:
            raise QiskitOptimizationError("Lowerbound is greater than upperbound!")
        self._upperbound = upperbound

    @property
    def vartype(self) -> VarType:
        """Returns the type of the variable.

        Returns:
            The variable type.

        """
        return self._vartype

    @vartype.setter
    def vartype(self, vartype: VarType) -> None:
        """Sets the type of the variable.

        Args:
            vartype: The variable type.
        """
        self._vartype = vartype

    def as_tuple(self) -> Tuple[str, Union[float, int], Union[float, int], VarType]:
        """Returns a tuple corresponding to this variable.

        Returns:
            A tuple corresponding to this variable consisting of name, lowerbound, upperbound and
            variable type.
        """
        return self.name, self.lowerbound, self.upperbound, self.vartype

    def __repr__(self):
        return f"<{self.__class__.__name__}: {str(self)}>"

    def __str__(self):
        if self._vartype == VarType.BINARY:
            return f"{self.name} ({self._vartype.name.lower()})"
        lowerbound = "" if self._lowerbound == -INFINITY else f"{self._lowerbound} <= "
        upperbound = "" if self._upperbound == INFINITY else f" <= {self._upperbound}"
        return f"{lowerbound}{self.name}{upperbound} ({self._vartype.name.lower()})"



from q3as.client import Client, Credentials
from q3as.algo.vqe import VQE
from q3as.run_options import RunOptions, EstimatorOptions, SamplerOptions
from q3as.api import Job, JobRequest, JobStatus, BaseJob

__all__ = [
    "VQE",
    "Client",
    "Credentials",
    "RunOptions",
    "EstimatorOptions",
    "SamplerOptions",
    "Job",
    "BaseJob",
    "JobRequest",
    "JobStatus",
]



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Interface for all objects that have a parent QuadraticProgram."""

# We import `problems` module not `QuadraticProgram` class
# to resolve the circular import issue of sphinx.
# See https://github.com/agronholm/sphinx-autodoc-typehints#dealing-with-circular-imports

from q3as.quadratic import problems  # pylint: disable=unused-import, cyclic-import


class QuadraticProgramElement:
    """Interface class for all objects that have a parent QuadraticProgram."""

    def __init__(self, quadratic_program: "problems.QuadraticProgram") -> None:
        """Initialize object with parent QuadraticProgram.

        Args:
            quadratic_program: The parent QuadraticProgram.
        Raises:
            TypeError: QuadraticProgram instance expected.
        """
        # pylint: disable=cyclic-import
        from .quadratic_program import QuadraticProgram

        if not isinstance(quadratic_program, QuadraticProgram):
            raise TypeError("QuadraticProgram instance expected")

        self._quadratic_program = quadratic_program

    @property
    def quadratic_program(self) -> "problems.QuadraticProgram":
        """Returns the parent QuadraticProgram.

        Returns:
            The parent QuadraticProgram.
        """
        return self._quadratic_program

    @quadratic_program.setter
    def quadratic_program(self, quadratic_program: "problems.QuadraticProgram") -> None:
        """Sets the parent QuadraticProgram.

        Args:
            quadratic_program: The parent QuadraticProgram.
        Raises:
            TypeError: QuadraticProgram instance expected.
        """
        # pylint: disable=cyclic-import
        from .quadratic_program import QuadraticProgram

        if not isinstance(quadratic_program, QuadraticProgram):
            raise TypeError("QuadraticProgram instance expected")

        self._quadratic_program = quadratic_program



from __future__ import annotations
from typing import cast, Dict, Any, List, Union, Mapping
from enum import Enum
import io
import base64

from pydantic import BaseModel
from qiskit.circuit import QuantumCircuit
from qiskit.primitives import PrimitiveResult, DataBin, BitArray
from qiskit.primitives.containers.observables_array import (
    ObservablesArray,
    ObservablesArrayLike,
)
from qiskit.primitives.containers.pub_result import PubResult
from qiskit.primitives.containers.sampler_pub_result import SamplerPubResult
import qiskit.qpy

from .numpy import EncodedArray

from q3as.algo.types import EstimatorData, SamplerData


class QuantumCircuitEncoding(Enum):
    QPY = "qpy"


class EncodedQuantumCircuit(BaseModel):
    encoding: QuantumCircuitEncoding
    base64: bytes

    @classmethod
    def encode(cls, qc: QuantumCircuit) -> EncodedQuantumCircuit:
        data = io.BytesIO()
        qiskit.qpy.dump(qc, data)
        return cls(
            encoding=QuantumCircuitEncoding.QPY,
            base64=base64.b64encode(data.getvalue()),
        )

    def decode(self) -> QuantumCircuit:
        if self.encoding != QuantumCircuitEncoding.QPY:
            raise ValueError("Only QPY encoding is supported")
        data = base64.b64decode(self.base64)
        loaded = qiskit.qpy.load(io.BytesIO(data))
        if len(loaded) != 1:
            raise ValueError("Expected single quantum circuit")
        if not isinstance(loaded[0], QuantumCircuit):
            raise ValueError("Expected quantum circuit")
        return loaded[0]


type EncodedObservable = Union[Mapping[str, float], List[EncodedObservable]]


class EncodedObservables(BaseModel):
    data: EncodedObservable

    @classmethod
    def encode(cls, obs: ObservablesArrayLike) -> EncodedObservables:
        return EncodedObservables(data=ObservablesArray.coerce(obs).tolist())

    def decode(self) -> ObservablesArray:
        return ObservablesArray(cast(ObservablesArrayLike, self.data))


class EncodedBitArray(BaseModel):
    array: EncodedArray
    num_bits: int

    @classmethod
    def encode(cls, array: BitArray) -> EncodedBitArray:
        return cls(array=EncodedArray.encode(array.array), num_bits=array.num_bits)

    def decode(self) -> BitArray:
        return BitArray(array=self.array.decode(), num_bits=self.num_bits)


class EncodedEstimatorPubResult(BaseModel):
    metadata: Dict[str, Any]
    evs: float
    stds: float

    @classmethod
    def encode(cls, result: PubResult) -> EncodedEstimatorPubResult:
        data = cast(EstimatorData, result.data)
        return cls(metadata=result.metadata, evs=data.evs, stds=data.stds)

    def decode(self) -> PubResult:
        return PubResult(
            metadata=self.metadata, data=DataBin(evs=self.evs, stds=self.stds)
        )


class EncodedSamplerPubResult(BaseModel):
    metadata: Dict[str, Any]
    meas: EncodedBitArray

    @classmethod
    def encode(cls, result: SamplerPubResult) -> EncodedSamplerPubResult:
        data = cast(SamplerData, result.data)
        return cls(metadata=result.metadata, meas=EncodedBitArray.encode(data.meas))

    def decode(self) -> SamplerPubResult:
        return SamplerPubResult(
            metadata=self.metadata, data=DataBin(meas=self.meas.decode())
        )


class EncodedEstimatorResult(BaseModel):
    metadata: Dict[str, Any]
    pub_results: List[EncodedEstimatorPubResult]

    @classmethod
    def encode(cls, result: PrimitiveResult[PubResult]) -> EncodedEstimatorResult:
        return cls(
            metadata=result.metadata,
            pub_results=[EncodedEstimatorPubResult.encode(r) for r in iter(result)],
        )

    def decode(self) -> PrimitiveResult[PubResult]:
        return PrimitiveResult(
            pub_results=[r.decode() for r in self.pub_results],
            metadata=self.metadata,
        )


class EncodedSamplerResult(BaseModel):
    metadata: Dict[str, Any]
    pub_results: List[EncodedSamplerPubResult]

    @classmethod
    def encode(cls, result: PrimitiveResult[SamplerPubResult]) -> EncodedSamplerResult:
        return cls(
            metadata=result.metadata,
            pub_results=[EncodedSamplerPubResult.encode(r) for r in iter(result)],
        )

    def decode(self) -> PrimitiveResult[SamplerPubResult]:
        return PrimitiveResult(
            pub_results=[r.decode() for r in self.pub_results],
            metadata=self.metadata,
        )



import os

from q3as.algo.vqe import VQE, EfficientSU2
from q3as.quadratic import QuadraticProgram
from q3as.app import Qubo
from q3as import Client, Credentials


from qiskit_aer import AerSimulator
from qiskit_aer.primitives import EstimatorV2 as Estimator, SamplerV2 as Sampler

from qiskit import transpile

url = os.getenv("Q3AS_URL", "http://localhost:8080")

simulator = AerSimulator()


def simple_qp():
    qp = QuadraticProgram("qp")

    qp.integer_var(name="x", lowerbound=-1, upperbound=1)
    qp.integer_var(name="y", lowerbound=-1, upperbound=5)
    qp.integer_var(name="z", lowerbound=-1, upperbound=5)
    qp.minimize(constant=3, linear={"x": 1}, quadratic={("x", "y"): 2, ("z", "z"): -1})

    return qp


def test_quadratic():
    _ = (
        VQE.builder()
        .app(Qubo(simple_qp()))
        .ansatz(EfficientSU2(lambda c: transpile(c, simulator)))
        .run(Estimator(), Sampler())
    )


def test_client_quadratic():
    with Client(Credentials.load(".credentials.json"), url=url) as client:
        _ = VQE.builder().app(Qubo(simple_qp())).send(client)



from enum import Enum

from qiskit.primitives import BitArray


class HaltReason(Enum):
    """
    The reason for halting
    """

    TOLERANCE = "tolerance"
    "The tolerance was reached"
    INTERRUPT = "interrupt"
    "The process was interrupted"
    MAXITER = "maxiter"
    "The maximum number of iterations was reached"


class EstimatorData:
    evs: float
    stds: float


class SamplerData:
    meas: BitArray
